/* 
   cmds.c -- handles:
   commands from a user via dcc

   dprintf'ized, 3nov95
 */
/*
   This file is part of the eggdrop source code
   copyright (c) 1997 Robey Pointer
   and is distributed according to the GNU general public license.
   For full details, read the top of 'main.c' or the file called
   COPYING that was distributed with this code.
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include "eggdrop.h"
#include "users.h"
#include "chan.h"
#include "tclegg.h"

extern int serv;
extern char botuserhost[];
extern char newserver[];
extern char newserverpass[];
extern int newserverport;
extern int require_p;
extern int use_info;
extern char origbotname[];
extern char botnetnick[];
extern int dcc_total;
extern struct dcc_t dcc[];
extern char botname[];
extern char botuser[];
extern char dccdir[];
extern char hostfile[];
extern struct userrec *userlist;
extern char owner[];
extern struct chanset_t *chanset;
extern int make_userfile;
extern int conmask;
extern Tcl_Interp *interp;
extern tcl_timer_t *timer, *utimer;
extern char chanfile[];
extern int gban_total;
extern int do_restart;
extern int default_port;
extern int noshare;
extern int backgrd;

/* temporary variables for the plus and minus telnet host routines */
/* Sorry for the untidyness, but it works :) */
char tmp[255];
char tmp2[255];
char tmp3[255];
int tmp4;

FILE *file;
FILE *file2;

/* include the chopped out bits if no modules */
#ifndef MODULES
#include "mod/filesys.mod/filesys.c"
#endif

/* Do we have any flags that will allow us ops on a channel? */
int has_op PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   int atr, chatr;
   context;
   if (par[0]) {
      chan = findchan(par);
      if (chan == NULL) {
      dprintf(idx," [1;32m No such channel.[0m\n");	 
      	 return 0;
      }
   } else {
      chan = findchan(dcc[idx].u.chat->con_chan);
      if (chan == NULL) {
      dprintf(idx," [1;34mInvalid console channel.[0m\n");	 
      	 return 0;
      }
   }
   atr = get_attr_handle(dcc[idx].nick);
   chatr = get_chanattr_handle(dcc[idx].nick, chan->name);
   if ((atr & (USER_MASTER | USER_OWNER)) ||
       (chatr & (CHANUSER_MASTER | CHANUSER_OWNER | CHANUSER_OP)))
      return 1;
   else if ((atr & USER_GLOBAL) && !(chatr & CHANUSER_DEOP))
      return 1;
   else {
     dprintf(idx," [1;33mYour are not a channel op on %s.[0m\n", chan->name); 
      return 0;
   }
   return 1;
}

void cmd_voice PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   char nick[512];
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: voice <nick> [channel][0m\n");     
      return;
   }
   nsplit(nick, par);
   if (par[0]) {
      if (!has_op(idx, par))
	 return;
      chan = findchan(par);
   } else {
      if (!has_op(idx, ""))
	 return;
      chan = findchan(dcc[idx].u.chat->con_chan);
   }
   if (!me_op(chan)) {
     dprintf(idx,"[1;34mI can't help you now because I'm not a chan op on %s.[0m\n",
     	      chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m voice (%s) %s %s",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan, nick, par);
   give_voice(nick, chan, idx);
} 

void cmd_devoice PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   char nick[512];
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: devoice <nick> [channel][0m\n"); 
      return;
   }
   nsplit(nick, par);
   if (par[0]) {
      if (!has_op(idx, par))
	 return;
      chan = findchan(par);
   } else {
      if (!has_op(idx, ""))
	 return;
      chan = findchan(dcc[idx].u.chat->con_chan);
   }
   if (!me_op(chan)) {
    dprintf(idx,"[1;34mI can't help you now because I'm not a chan op on %s.[0m\n", 
    	      chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) devoice %s %s",dcc[idx].nick,
	  dcc[idx].u.chat->con_chan, nick, par);
   give_devoice(nick, chan, idx);
}

void cmd_who PROTO2(int, idx, char *, par)
{
   int i;
   if (par[0]) {
      if (dcc[idx].u.chat->channel < 0) {
      dprintf(idx," [1;34m You have chat turned off.[0m\n");	 
      	 return;
      }
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m who %s",dcc[idx].nick,par);
      if (strcasecmp(par, botnetnick) == 0)
	 tell_who(idx, dcc[idx].u.chat->channel);
      else {
	 i = nextbot(par);
	 if (i < 0) {
	 dprintf(idx,"[1;37mThat bot isn't connected.[0m\n");   
/*      tandout("who %d:%s@%s %s %d\n",dcc[idx].sock,dcc[idx].nick,
   botnetnick,par,dcc[idx].u.chat->channel);  */
	 } else if (dcc[idx].u.chat->channel > 99999)
	  dprintf(idx,"[1;34mYou are on a local channel[0m\n");  
	 else
	    tprintf(dcc[i].sock, "who %d:%s@%s %s %d\n", dcc[idx].sock,
	       dcc[idx].nick, botnetnick, par, dcc[idx].u.chat->channel);
      }
   } else {
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m who",dcc[idx].nick);   
         if (dcc[idx].u.chat->channel < 0)
	 tell_who(idx, 0);
      else
	 tell_who(idx, dcc[idx].u.chat->channel);
   }
}

void cmd_botinfo PROTO2(int, idx, char *, par)
{
 putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m botinfo",dcc[idx].nick);  
   tandout("info? %d:%s@%s\n", dcc[idx].sock, dcc[idx].nick, botnetnick);
}

void cmd_whom PROTO2(int, idx, char *, par)
{
   if (dcc[idx].u.chat->channel < 0) {
      dprintf(idx,"[1;34mYou have chat turned off.[0m\n");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m whom %s",dcc[idx].nick,par);
   if (!par[0])
      answer_local_whom(idx, dcc[idx].u.chat->channel);
   else if (par[0] == '*')
      answer_local_whom(idx, -1);
   else {
      int chan = atoi(par);
      if (par[0] == '*') {
	 if ((par[1] < '0' || par[1] > '9')) {
	    chan = get_assoc(par + 1);
	    if (chan < 0) {
	     dprintf(idx,"[1;34mNo such channel.[0m\n");  
	       return;
	    }
	 } else
	    chan = 100000 + atoi(par + 1);
	 if (chan < 100000 || chan > 199999) {
	  dprintf(idx,"[1;34mChannel # out of range: must be 0-99999[0m\n");  
	     return;
	 }
      } else {
	 if ((par[0] < '0') || (par[0] > '9')) {
	    chan = get_assoc(par);
	    if (chan < 0) {
	    dprintf(idx,"[1;34mNo such channel.[0m\n");   
	       return;
	    }
	 }
	 if ((chan < 0) || (chan > 99999)) {
	 dprintf(idx,"[1;34mChannel # out of range: must be 0-99999[0m\n");   
	     return;
	 }
      }
      answer_local_whom(idx, chan);
   }
}

void cmd_me PROTO2(int, idx, char *, par)
{
   int i;
   if (dcc[idx].u.chat->channel < 0) {
   dprintf(idx,"[1;34mYou have chat turned off.[0m\n");   
         return;
   }
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage: me <action>[0m\n");     
         return;
   }
   if (dcc[idx].u.chat->away != NULL)
      not_away(idx);
   for (i = 0; i < dcc_total; i++) {
      if ((dcc[i].type == DCC_CHAT) &&
	  (dcc[i].u.chat->channel == dcc[idx].u.chat->channel) &&
	  ((i != idx) || (dcc[i].u.chat->status & STAT_ECHO)))
	 dprintf(i,"[1;36m*[0m %s %s\n",dcc[idx].nick,par);
	  if (dcc[i].type == DCC_BOT)
	 if (dcc[idx].u.chat->channel < 100000)
	    tprintf(dcc[i].sock, "actchan %s@%s %d %s\n", dcc[idx].nick, botnetnick,
		    dcc[idx].u.chat->channel, par);
   }
   check_tcl_act(dcc[idx].nick, dcc[idx].u.chat->channel, par);
}

void cmd_motd PROTO2(int, idx, char *, par)
{
   int i;
   if (par[0]) {
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m motd %s",dcc[idx].nick,par);   
         if (strcasecmp(par, botnetnick) == 0)
	 show_motd(idx);
      else {
	 i = nextbot(par);
	 if (i < 0)
	 dprintf(idx,"[1;34mThat bot isn't connected.[0m\n");   
	 else
	    tprintf(dcc[i].sock, "motd %d:%s@%s %s\n", dcc[idx].sock,
		    dcc[idx].nick, botnetnick, par);
      }
   } else {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m motd",dcc[idx].nick);
      show_motd(idx);
   }
}

void cmd_note PROTO2(int, idx, char *, par)
{
   char handle[512], *p;
   int echo;
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: note <to-whom> <message>[0m\n");     
         return;
   }
   /* could be file system user */
   nsplit(handle, par);
   echo = (dcc[idx].type == DCC_CHAT) ? (dcc[idx].u.chat->status & STAT_ECHO) :
       (dcc[idx].u.file->chat->status & STAT_ECHO);
   p = strchr(handle, ',');
   while (p != NULL) {
      *p = 0;
      p++;
      add_note(handle, dcc[idx].nick, par, idx, echo);
      strcpy(handle, p);
      p = strchr(handle, ',');
   }
   add_note(handle, dcc[idx].nick, par, idx, echo);
}

void cmd_away PROTO2(int, idx, char *, par)
{
   if (strlen(par) > 60)
      par[60] = 0;
   set_away(idx, par);
}

void cmd_newpass PROTO2(int, idx, char *, par)
{
   char new[512];
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: newpass <newpassword>[0m\n");   
         return;
   }
   nsplit(new, par);
   if (strlen(new) > 16)
      new[16] = 0;
   if (strlen(new) < 4) {
      dprintf(idx,"[1;34mPlease use at least 4 characters.[0m\n");
      return;
   }
   change_pass_by_handle(dcc[idx].nick, new);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m newpass...",dcc[idx].nick);
   dprintf(idx,"[1;34mChanged password to[0m [1;31m'%s'[0m[1;34m[0m\n",new);
}

void cmd_bots PROTO2(int, idx, char *, par)
{
  putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m bots",dcc[idx].nick); 
   tell_bots(idx);
}

void cmd_bottree PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m bottree",dcc[idx].nick);
   tell_bottree(idx);
}

void cmd_help PROTO2(int, idx, char *, par)
{
   int atr, chatr;
   atr = get_attr_handle(dcc[idx].nick);
   chatr = get_chanattr_handle(dcc[idx].nick, dcc[idx].u.chat->con_chan);
   if (chatr & CHANUSER_OP)
      atr |= USER_PSUEDOOP;
   if (chatr & CHANUSER_MASTER)
      atr |= USER_PSUMST;
   if (chatr & CHANUSER_OWNER)
      atr |= USER_PSUOWN;
   if (par[0]) {
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m help %s",dcc[idx].nick,par);  
         tellhelp(idx, par, atr);
   } else {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m help",dcc[idx].nick);
      if (atr & (USER_GLOBAL | USER_PSUEDOOP | USER_MASTER | USER_PSUMST | USER_BOTMAST))
	 tellhelp(idx, "help", atr);
      else
	 tellhelp(idx, "helpparty", atr);
   }
}

#ifndef NO_IRC
void cmd_act PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: act <action>[0m\n");  
      return;
   }
   if (!has_op(idx, ""))
      return;
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m(%s) act %s",dcc[idx].nick,
	  dcc[idx].u.chat->con_chan, par);
   mprintf(serv, "PRIVMSG %s :\001ACTION %s\001\n", dcc[idx].u.chat->con_chan,
	   par);
}
#endif

#ifndef NO_IRC
void cmd_msg PROTO2(int, idx, char *, par)
{
   char nick[512];
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage: msg <nick> <message>[0m\n");
   } else {
      nsplit(nick, par);
     putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m msg %s %s",dcc[idx].nick,nick,par);
      mprintf(serv, "PRIVMSG %s :%s\n", nick, par);
   }
}
#endif

#ifndef NO_IRC
void cmd_say PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: say <message>[0m\n");  
         return;
   }
   if (!has_op(idx, ""))
      return;
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m say %s",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan, par);
   mprintf(serv, "PRIVMSG %s :%s\n", dcc[idx].u.chat->con_chan, par);
}
#endif

#ifndef NO_IRC
void cmd_kickban PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: kickban <nick> [reason][0m\n");   
         return;
   }
   if (!has_op(idx, ""))
      return;
   chan = findchan(dcc[idx].u.chat->con_chan);
   if (!me_op(chan)) {
      dprintf(idx,"[1;34mI can't help you now because I'm not a channel on %s.[0m\n",
      	      chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) kickban %s",dcc[idx].nick,
	  dcc[idx].u.chat->con_chan, par);
   user_kickban(idx, par);
}
#endif

#ifndef NO_IRC
void cmd_op PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   char nick[512];
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: op <nick> [channel][0m\n");     
      return;
   }
   nsplit(nick, par);
   if (par[0]) {
      if (!has_op(idx, par))
	 return;
      chan = findchan(par);
   } else {
      if (!has_op(idx, ""))
	 return;
      chan = findchan(dcc[idx].u.chat->con_chan);
   }
   if (!me_op(chan)) {
     dprintf(idx,"[1;34mI can't help you now because I'm not a chan op on %s.[0m\n",
     	      chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m op (%s) %s %s",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan, nick, par);
   give_op(nick, chan, idx);
}
#endif

#ifndef NO_IRC
void cmd_deop PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   char nick[512];
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: deop <nick> [channel][0m\n"); 
      return;
   }
   nsplit(nick, par);
   if (par[0]) {
      if (!has_op(idx, par))
	 return;
      chan = findchan(par);
   } else {
      if (!has_op(idx, ""))
	 return;
      chan = findchan(dcc[idx].u.chat->con_chan);
   }
   if (!me_op(chan)) {
    dprintf(idx,"[1;34mI can't help you now because I'm not a chan op on %s.[0m\n", 
    	      chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) deop %s %s",dcc[idx].nick,
	  dcc[idx].u.chat->con_chan, nick, par);
   give_deop(nick, chan, idx);
}
#endif

#ifndef NO_IRC
void cmd_kick PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: kick <nick> [reason][0m\n");  
         return;
   }
   if (!has_op(idx, ""))
      return;
   chan = findchan(dcc[idx].u.chat->con_chan);
   if (!me_op(chan)) {
      dprintf(idx,"[1;34mI can't help you now because I'm not a channel op on %s.[0m\n",
	      chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) kick %s",dcc[idx].nick,
	  dcc[idx].u.chat->con_chan, par);
   user_kick(idx, par);
}
#endif

#ifndef NO_IRC
void cmd_invite PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   if (!has_op(idx, ""))
      return;
   chan = findchan(dcc[idx].u.chat->con_chan);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) invite %s",dcc[idx].nick,
  	  dcc[idx].u.chat->con_chan, par);
   if (!me_op(chan) && (chan->channel.mode & CHANINV)) {
    dprintf(idx,"[1;34mI'm not chop, so I can't invite anyone.[0m\n");  
         return;
   }
   if (ischanmember(chan->name, par) && !is_split(chan->name, par)) {
     dprintf(idx,"[1;34m%s is already on %s![0m\n",par, chan->name); 
      return;
   }
   mprintf(serv, "INVITE %s %s\n", par, chan->name);
   dprintf(idx,"[1;34mInviting %s to %s.[0m\n",par,chan->name);
}
#endif

void cmd_resetbans PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   if (!has_op(idx, ""))
      return;
   chan = findchan(dcc[idx].u.chat->con_chan);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) resetbans",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan);
   dprintf(idx,"[1;34mResetting bans on %s...[0m\n",chan->name);
   resetbans(chan);
}

void cmd_pls_ban PROTO2(int, idx, char *, par)
{
   char who[512], note[512], s[UHOSTLEN];
   struct chanset_t *chan;
   char chname[512];
   chname[0] = 0;
   if (!par[0]) {
    dprintf(idx,"[1l34mUsage: +ban <hostmask> <channel> [reason][0m\n");  
         return;
   }
   nsplit(who, par);
   rmspace(who);
   chan = findchan(dcc[idx].u.chat->con_chan);
   if ((par[0] == '#') || (par[0] == '&')) {
      nsplit(chname, par);
      rmspace(chname);
      if (!has_op(idx, chname))
	 return;
      chan = findchan(chname);
   } else if (!(get_attr_handle(dcc[idx].nick) & USER_MASTER)) {
      if (!has_op(idx, ""))
	 return;
      strcpy(chname, dcc[idx].u.chat->con_chan);
   }
   rmspace(par);
   if (!par[0])
      strcpy(note, "request");
   else {
      strncpy(note, par, 65);
      note[65] = 0;
   }
   /* fix missing ! or @ BEFORE checking against myself */
   if (strchr(who, '!') == NULL)
      strcat(s, "!*@*");	/* lame nick ban */
   if (strchr(who, '@') == NULL)
      strcat(s, "@*");		/* brain-dead? */
   sprintf(s, "%s!%s", botname, botuserhost);
   if (wild_match(who, s)) {
      dprintf(idx,"[1;34mI'm gonna ban myself.. heh.. (=[0m\n");
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m attempted +ban %s",dcc[idx].nick,who);
      return;
   }
   if (strlen(who) > 70)
      who[70] = 0;
   /* irc can't understand bans longer than that */
   if (chname[0]) {
      u_addban(chan->bans, who, dcc[idx].nick, note, 0L);
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) +ban %s %s (%s)",dcc[idx].nick,
	     dcc[idx].u.chat->con_chan, who, chname, note);
      dprintf(idx,"[1;34mNew %s ban: %s (%s)[0m\n", chname, who, note);
      if (me_op(chan))
	 add_mode(chan, '+', 'b', who);
      recheck_channel(chan);
      return;
   }
   addban(who, dcc[idx].nick, note, 0L);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) +ban %s (%s)",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan, who, note);
   dprintf(idx,"[1;34mNew ban: %s (%s)[0m\n",who,note);
   chan = chanset;
   while (chan != NULL) {
      if (me_op(chan))
	 add_mode(chan, '+', 'b', who);
      recheck_channel(chan);
      chan = chan->next;
   }
}

void cmd_mns_ban PROTO2(int, idx, char *, par)
{
   int i = 0, j, k;
   struct chanset_t *chan;
   char s[UHOSTLEN], ban[512], chname[512];
   chname[0] = 0;
   if (!par[0]) {
    dprintf(idx,"[1l34mUsage: -ban <hostmask|ban #> [channel][0m\n");  
         return;
   }
   nsplit(ban, par);
   rmspace(ban);
   rmspace(par);
   if (par[0] == '#') {
      strcpy(chname, par);
      if (!has_op(idx, chname))
	 return;
   } else if (!has_op(idx, ""))
      return;
   if (!chname[0] && (get_attr_handle(dcc[idx].nick) & USER_MASTER)) {
      i = delban(ban);
      if (i > 0) {
         putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) -ban %s",dcc[idx].nick,
		dcc[idx].u.chat->con_chan, ban);
	 dprintf(idx,"[1;34mRemoved ban:[0m [1;50m%s[0m\n",ban);
	 chan = chanset;
	 while (chan != NULL) {
	    if (me_op(chan))
	       add_mode(chan, '-', 'b', ban);
	    chan = chan->next;
	 }
	 return;
      }
   }
   /* channel-specific ban? */
   if (!chname[0])
      strcpy(chname, dcc[idx].u.chat->con_chan);
   chan = findchan(chname);
   sprintf(s, "%d", i + atoi(ban));
   j = u_delban(chan->bans, s);
   if (j > 0) {
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) -ban %s",dcc[idx].nick,   
  	     dcc[idx].u.chat->con_chan, s);
      dprintf(idx,"[1;34mRemoved[0m [1;50m%s[0m [1;31mchannel ban:[0m [1;33m%s[0m\n", chname, s);
      add_mode(chan, '-', 'b', s);
      return;
   }
   k = u_delban(chan->bans, ban);
   if (k > 0) {
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) -ban %s",dcc[idx].nick,   
   	     dcc[idx].u.chat->con_chan, ban);
     dprintf(idx,"[1;34mRemoved[0m [1;50m%s[0m [1;31mchannel ban:[0m [1;33m%s[0m\n", chname, ban);
           add_mode(chan, '-', 'b', ban);
      return;
   }
   /* okay, not in any ban list -- might be ban on channel */
   if (atoi(ban) > 0) {
      if (kill_chanban(chname, idx, 1 - i - j, atoi(ban))) {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) -ban %s [on channel]",	 
     		dcc[idx].nick, dcc[idx].u.chat->con_chan, ban);
	 dprintf(idx,"[1;34mNot in banlist. Removing match on %s[0m\n", chname);
	 return;
      }
   } else {
      if (kill_chanban_name(chname, idx, ban)) {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) -ban %s [on channel]",	 
      		dcc[idx].nick, dcc[idx].u.chat->con_chan, ban);
	 dprintf(idx,"[1;34mNot in banlist. Removing match on[0m [1;33m%s[0m\n", chname);
	 return;
      }
   }
}

void cmd_bans PROTO2(int, idx, char *, par)
{
   if (strcasecmp(par, "all") == 0) {
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m bans all",dcc[idx].nick);  
         tell_bans(idx, 1, "");
   } else {
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m bans %s",dcc[idx].nick,par); 
      tell_bans(idx, 0, par);
   }
}

#ifndef NO_IRC
void cmd_channel PROTO2(int, idx, char *, par)
{
   if (par[0]) {
      if (!has_op(idx, par))
	 return;
   } else if (!has_op(idx, ""))
      return;
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) channel %s",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan, par);
   tell_verbose_chan_info(idx, par);
}
#endif

void cmd_addlog PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: addlog <message>[0m\n"); 
      return;
   }
   dprintf(idx,"[1;34mPlaced entry in the log file.[0m\n");
   putlog(LOG_MISC, "*", "%s: %s", dcc[idx].nick, par);
}

#ifndef NO_IRC
void cmd_servers PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m servers",dcc[idx].nick);
   tell_servers(idx);
}

#endif

void cmd_whois PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: whois <handle>[0m\n");   
         return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m whois %s",dcc[idx].nick,par);
   tell_user_ident(idx, par, (get_attr_handle(dcc[idx].nick) & USER_MASTER));
}

void cmd_match PROTO2(int, idx, char *, par)
{
   int start = 1, limit = 20;
   char s[512], s1[512], chname[512];
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: match <nick/host>[0m\n");   
         return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m match %s",dcc[idx].nick,par);
   nsplit(s, par);
   if ((par[0] == '#') || (par[0] == '+') || (par[0] == '&'))
      nsplit(chname, par);
   else
      chname[0] = 0;
   if (atoi(par) > 0) {
      split(s1, par);
      if (atoi(s1) > 0)
	 start = atoi(s1);
      limit = atoi(par);
   }
   tell_users_match(idx, s, start, limit, (get_attr_handle(dcc[idx].nick) &
					   USER_MASTER), chname);
}

void cmd_status PROTO2(int, idx, char *, par)
{
   int atr = 0;
   if (strcasecmp(par, "all") == 0) {
      atr = get_attr_handle(dcc[idx].nick);
      if (!(atr & USER_MASTER)) {
	dprintf(idx,"[1;34mYou do not have Bot Master privelages[0m\n");  
      	 return;
      }
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m status all",dcc[idx].nick);
      tell_verbose_status(idx, 1);
      tell_mem_status_dcc(idx);
      dprintf(idx, "\n");
      tell_settings(idx);
   } else {
     putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m status",dcc[idx].nick); 
        tell_verbose_status(idx, 1);
      tell_mem_status_dcc(idx);
   }
}

void cmd_dccstat PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m dccstat",dcc[idx].nick);
   tell_dcc(idx);
}

void cmd_pls_ignore PROTO2(int, idx, char *, par)
{
   char who[512], note[66];
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: +ignore <hostmask> [comment][0m\n"); 
      return;
   }
   nsplit(who, par);
   who[UHOSTLEN - 1] = 0;
   strncpy(note, par, 65);
   note[65] = 0;
   if (match_ignore(who)) {
    dprintf(idx,"[1;34mThat already matches an existing ignore.[0m\n");  
         return;
   }
   dprintf(idx,"[1;34mNow ignoring: %s (%s)[0m\n",who,note);
   addignore(who, dcc[idx].nick, note, 0L);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m +ignore %s %s",dcc[idx].nick,who,note);
}

void cmd_mns_ignore PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: -ignore <hostmask>[0m\n");  
         return;
   }
   if (delignore(par)) {
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m -ignore %s",dcc[idx].nick,par);
    dprintf(idx,"[1;34mNo longer ignoring: %s[0m\n",par);
 } else
    dprintf(idx,"[1;34mCan't find that ignore.[0m\n");     
}

void cmd_ignores PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m ignores %s",dcc[idx].nick,par);
   tell_ignores(idx, par);
}

void cmd_boot PROTO2(int, idx, char *, par)
{
   int i, files = 0, ok = 0;
   char who[512];
   if (!par[0]) {
      dprintf(idx, "Usage: boot nick[@bot]\n");
      return;
   }
   nsplit(who, par);
   if (strchr(who, '@') != NULL) {
      char whonick[512];
      splitc(whonick, who, '@');
      whonick[20] = 0;
      if (strcasecmp(who, botnetnick) == 0) {
	 cmd_boot(idx, whonick);
	 return;
      }
#ifdef REMOTE_BOOTS
      i = nextbot(who);
      if (i < 0) {
	dprintf(idx,"[1;34mNo such bot connected.[0m\n"); 
         return;
      }
      tprintf(dcc[i].sock, "reject %s@%s %s@%s %s\n", dcc[idx].nick, botnetnick,
	      whonick, who, par[0] ? par : dcc[idx].nick);
      putlog(LOG_MISC,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m boot %s@%s (%s)",dcc[idx].nick,whonick,who,
	     par[0] ? par : dcc[idx].nick);
#else
      dprintf(idx,"[1;34mRemote boots are disabled here.[0m\n");
#endif
      return;
   }
   for (i = 0; i < dcc_total; i++)
      if ((strcasecmp(dcc[i].nick, who) == 0) && (!ok) &&
	  ((dcc[i].type == DCC_CHAT) || (dcc[i].type == DCC_FILES))) {
#ifdef OWNER
	 if ((get_attr_handle(who) & USER_OWNER) &&
	     (strcasecmp(dcc[idx].nick, who) != 0)) {
	    dprintf(idx,"[1;34mCan't boot the bot owner, dumb ass![0m\n");
	    return;
	 }
#endif
	 if ((get_attr_handle(who) & USER_MASTER) &&
	     (!(get_attr_handle(dcc[idx].nick) & USER_MASTER))) {
	    dprintf(idx,"[1;34mCan't boot a bot master.[0m\n");
	    return;
	 }
	 files = (dcc[i].type == DCC_FILES);
	 if (files)
	 dprintf(idx,"[1;34mBooted %s from the file section.[0m\n",dcc[i].nick);   
	 else
	   dprintf(idx,"[1;34mBooted %s from the bot.[0m\n",dcc[i].nick); 
	 putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m boot %s %s",dcc[idx].nick,who,par);
	 do_boot(i, dcc[idx].nick, par);
	 ok = 1;
      }
   if (!ok)
    dprintf(idx,"[1;34mWho?  No such person on the party line.[0m\n");
}

void cmd_console PROTO2(int, idx, char *, par)
{
   char nick[512], s[2], s1[512];
   int dest = 0, i, ok = 0, pls, md, atr, chatr;
   if (!par[0]) {
   dprintf(idx,"[1;34mYour console is %s: %s ([1;32m%s[0m)[0m\n",dcc[idx].u.chat->con_chan,   
   	      masktype(dcc[idx].u.chat->con_flags),
	      maskname(dcc[idx].u.chat->con_flags));
      return;
   }
   atr = get_attr_handle(dcc[idx].nick);
   strcpy(s1, par);
   split(nick, par);
   if ((nick[0]) && (nick[0] != '#') && (nick[0] != '&') &&
       (atr & (USER_MASTER | USER_OWNER))) {
      for (i = 0; i < dcc_total; i++)
	 if ((strcasecmp(nick, dcc[i].nick) == 0) && (dcc[i].type == DCC_CHAT) && (!ok)) {
	    ok = 1;
	    dest = i;
	 }
      if (!ok) {
      dprintf(idx,"[1;34mNo such user on the party line![0m\n");	 
      	 return;
      }
      nick[0] = 0;
   } else
      dest = idx;
   if (!nick[0])
      nsplit(nick, par);
   while (nick[0]) {
      if ((nick[0] == '#') || (nick[0] == '&') || (nick[0] == '*')) {
	 if ((nick[0] != '*') && (findchan(nick) == NULL)) {
	 dprintf(idx,"[1;34mInvalid console channel: %s[0m\n",nick);   
	     return;
	 }
	 chatr = get_chanattr_handle(dcc[idx].nick, nick);
	 if ((!(chatr & (CHANUSER_OP | CHANUSER_MASTER)))
	     && (!(atr & USER_MASTER)) &&
	     !((atr & USER_GLOBAL) && !(chatr & CHANUSER_DEOP))) {
	    dprintf(idx,"[1;34mYou don't have op or master access to channel %s[0m\n",nick);
	    return;
	 }
	 strncpy(dcc[dest].u.chat->con_chan, nick, 80);
	 dcc[dest].u.chat->con_chan[80] = 0;
      } else {
	 pls = 1;
	 if ((nick[0] != '+') && (nick[0] != '-'))
	    dcc[dest].u.chat->con_flags = 0;
	 for (i = 0; i < strlen(nick); i++) {
	    if (nick[i] == '+')
	       pls = 1;
	    else if (nick[i] == '-')
	       pls = (-1);
	    else {
	       s[0] = nick[i];
	       s[1] = 0;
	       md = logmodes(s);
	       if ((dest == idx) && !(atr & USER_MASTER) && pls) {
		  if (get_chanattr_handle(dcc[idx].nick, dcc[dest].u.chat->con_chan)
		      & CHANUSER_MASTER)
		     md &= ~(LOG_RAW | LOG_FILES | LOG_LEV1 | LOG_LEV2 |
			     LOG_LEV3 | LOG_LEV4 | LOG_LEV5 | LOG_LEV6 |
			     LOG_LEV7 | LOG_LEV8 | LOG_DEBUG);
		  else
		     md &= ~(LOG_MISC | LOG_CMDS | LOG_RAW | LOG_FILES | LOG_LEV1 | LOG_LEV2 |
			     LOG_LEV3 | LOG_LEV4 | LOG_LEV5 | LOG_LEV6 | LOG_LEV7 | LOG_LEV8 |
			     LOG_WALL | LOG_DEBUG);
	       }
	       if (pls == 1)
		  dcc[dest].u.chat->con_flags |= md;
	       else
		  dcc[dest].u.chat->con_flags &= ~md;
	    }
	 }
      }
      nsplit(nick, par);
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m console %s",dcc[idx].nick,s1);
   if (dest == idx) {
      dprintf(idx,"[1;31mSet your console to %s: %s ([1;32m%s[0m)[0m\n",
      	      dcc[idx].u.chat->con_chan,
	      masktype(dcc[idx].u.chat->con_flags),
	      maskname(dcc[idx].u.chat->con_flags));
   } else {
   dprintf(idx,"[1;34mSet console of %s to %s: %s ([1;32m%s[0m)[0m\n",dcc[dest].nick,   
   	      dcc[dest].u.chat->con_chan,
	      masktype(dcc[dest].u.chat->con_flags),
	      maskname(dcc[dest].u.chat->con_flags));
      dprintf(dest,"[1;33m%s set your console to %s: %s ([1;32m%s[0m)[0m\n",dcc[idx].nick,
      	      dcc[dest].u.chat->con_chan,
	      masktype(dcc[dest].u.chat->con_flags),
	      maskname(dcc[dest].u.chat->con_flags));
   }
}

#ifndef NO_IRC
void cmd_adduser PROTO2(int, idx, char *, par)
{
   char nick[512];
   struct chanset_t *chan;
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: adduser <nick> [handle][0m\n");  
         return;
   }
   chan = findchan(dcc[idx].u.chat->con_chan);
   if (chan == NULL) {
      dprintf(idx,"[1;34mYour console channel is invalid[0m\n");
      return;
   }
   nsplit(nick, par);
   if (!par[0]) {
      if (add_chan_user(nick, idx, nick))
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m adduser %s",dcc[idx].nick,nick);	 
} else {
      if (add_chan_user(nick, idx, par))
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m adduser %s %s",dcc[idx].nick,nick,par);	
   }
}
#endif

void cmd_pls_user PROTO2(int, idx, char *, par)
{
   char handle[512], host[512];
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: +user <handle> <hostmask>[0m\n");  
         return;
   }
   nsplit(handle, par);
   nsplit(host, par);
   if (strlen(handle) > 9)
      handle[9] = 0;		/* max len = 9 */
   if (is_user(handle)) {
     dprintf(idx,"[1;34mSomeone already exists by that name.[0m\n"); 
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m +user %s %s",dcc[idx].nick,handle,host);
   userlist = adduser(userlist, handle, host, "-", 0);
   dprintf(idx,"[1;34mAdded %s (%s) with no password or flags.[0m\n",handle,host);
}

void cmd_pls_bot PROTO2(int, idx, char *, par)
{
   char handle[512], addr[512];
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: +bot <handle> <address:port#>[0m\n");  
        return;
   }
   nsplit(handle, par);
   nsplit(addr, par);
   if (strlen(handle) > 9)
      handle[9] = 0;		/* max len = 9 */
   if (is_user(handle)) {
      dprintf(idx,"[1;34mSomeone already exists by that name.[0m\n");
      return;
   }
   if (strlen(addr) > 60)
      addr[60] = 0;
   if (strchr(addr, ':') == NULL)
      strcat(addr, ":3333");
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m +bot %s %s",dcc[idx].nick,handle,addr);
   userlist = adduser(userlist, handle, "none", "-", USER_BOT);
   set_handle_info(userlist, handle, addr);
   dprintf(idx,"[1;34mAdded bot '%s' with address '%s' and no password.[0m\n",
   	   handle, addr);
   if (!add_bot_hostmask(idx, handle))
     dprintf(idx,"[1;34mYou'll want to add a hostmask if this bot will ever %s[0m",
	    "[1;34mbe on any channels that I'm on.[0m\n"); 
     
}

#ifndef NO_IRC
void cmd_deluser PROTO2(int, idx, char *, par)
{
   char nick[512];
   struct chanset_t *chan;
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: deluser <nick>[0m\n");   
         return;
   }
   chan = findchan(dcc[idx].u.chat->con_chan);
   if (chan == NULL) {
     dprintf(idx,"[1;34mYour console channel is invalid.[0m\n"); 
      return;
   }
   nsplit(nick, par);
#ifdef OWNER
   if (!(get_attr_handle(dcc[idx].nick) & USER_OWNER) &&
       (get_attr_handle(par) & USER_OWNER)) {
      dprintf(idx,"[1;34mCan't remove the bot owner![0m\n");
      return;
   }
   if (!(get_chanattr_handle(dcc[idx].nick, chan->name) & CHANUSER_OWNER) &&
       (get_chanattr_handle(par, chan->name) & CHANUSER_OWNER) &&
       (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
      dprintf(idx,"[1;34mCan't remove the channel owner![0m\n");
      return;
   }
#endif
   if (del_chan_user(nick, idx))
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m deluser %s",dcc[idx].nick, nick); 
}
#endif

void cmd_mns_user PROTO2(int, idx, char *, par)
{
   int atr = get_attr_handle(dcc[idx].nick), atr1 = get_attr_handle(par);
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: -user <nick>[0m\n"); 
      return;
   }
#ifdef OWNER
   if (!(get_attr_handle(dcc[idx].nick) & USER_OWNER) &&
       (get_attr_handle(par) & USER_OWNER)) {
     dprintf(idx,"[1;34mCan't remove the bot owner![0m\n");
      return;
   }
   if ((get_attr_handle(par) & BOT_SHARE) &&
       (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
      dprintf(idx,"[1;34mCan't remove shared bots.[0m\n");
      return;
   }
#endif
   if ((atr & USER_BOTMAST) && (!(atr & (USER_MASTER | USER_OWNER)))
       && (!(atr1 & USER_BOT))) {
     dprintf(idx,"[1;34mCan't remove users who aren't bots![0m\n");
      return;
   }
   if (deluser(par)) {
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m -user %s",dcc[idx].nick,par);   
   dprintf(idx,"[1;34mDeleted %s.[0m\n",par);
 } else
  dprintf(idx,"[1;34mFailed. misspelled?[0m\n");   
}

void cmd_chnick PROTO2(int, idx, char *, par)
{
   char hand[512];
   int i;
   split(hand, par);
   if ((!hand[0]) || (!par[0])) {
   dprintf(idx,"[1;34mUsage: chnick <oldnick> <newnick>[0m\n");   
         return;
   }
   if (strlen(par) > 9)
      par[9] = 0;
   for (i = 0; i < strlen(par); i++)
      if ((par[i] <= 32) || (par[i] >= 127) || (par[i] == '@'))
	 par[i] = '?';
   if (strchr("-,+*=:!.@#;$", par[0]) != NULL) {
     dprintf(idx,"[1;34mBizarre quantum forces prevent nicknames from starting with %c[0m\n",par[0]); 
      return;
   }
   if ((is_user(par)) && (strcasecmp(hand, par) != 0)) {
   dprintf(idx,"[1;34mAlready a user %s.[0m\n",par);   
         return;
   }
   if ((strcasecmp(par, origbotname) == 0) || (strcasecmp(par, botnetnick) == 0)) {
      dprintf(idx,"[1;34mHey!  Thats MY name doof![0m\n",par);
      return;
   }
   if ((get_attr_handle(dcc[idx].nick) & USER_BOTMAST) &&
       (!(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_OWNER))) &&
       (!(get_attr_handle(hand) & USER_BOT))) {
       dprintf(idx,"[1;34mCan't change nicks for non-bots.[0m\n");
      return;
   }
#ifdef OWNER
   if ((get_attr_handle(hand) & BOT_SHARE) &&
       (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
      dprintf(idx,"[1;34mCan't change shared bot's nick.[0m\n");
      return;
   }
   if ((get_attr_handle(hand) & USER_OWNER) &&
       (strcasecmp(dcc[idx].nick, hand) != 0)) {
      dprintf(idx,"[1;34mCan't change the bot owner's handle.[0m\n");
      return;
   }
#endif
   if (change_handle(hand, par)) {
      notes_change(idx, hand, par);
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chnick %s %s",dcc[idx].nick,hand,par);
      dprintf(idx,"[1;34mChanged.[0m\n");
      for (i = 0; i < dcc_total; i++) {
	 if ((strcasecmp(dcc[i].nick, hand) == 0) && (dcc[i].type != DCC_BOT)) {
	    char s[10];
	    strcpy(s, dcc[i].nick);
	    strcpy(dcc[i].nick, par);
	    if ((dcc[i].type == DCC_CHAT) && (dcc[i].u.chat->channel >= 0)) {
	       chanout2(dcc[i].u.chat->channel, "Nick change: %s -> %s\n", s, par);
	       if (dcc[i].u.chat->channel < 100000) {
		  context;
		  tandout("part %s %s %d\n", botnetnick, s, dcc[i].sock);
		  tandout("join %s %s %d %c%d %s\n", botnetnick, par,
			  dcc[i].u.chat->channel, geticon(i), dcc[i].sock, dcc[i].host);
	       }
	    }
	 }
      }
   } else
    dprintf(idx,"[1;34mFailed.[0m\n");  
}


void cmd_nick PROTO2(int, idx, char *, par)
{
   int i;
   char icon;
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: nick <new-handle>[0m\n");  
         return;
   }
   if (strlen(par) > 9)
      par[9] = 0;
   for (i = 0; i < strlen(par); i++)
      if ((par[i] <= 32) || (par[i] >= 127) || (par[i] == '@'))
	 par[i] = '?';
   if (strchr("-,+*=:!.@#;$", par[0]) != NULL) {
   dprintf(idx,"[1;34mBizarre quantum forces prevent nicknames from starting with %c[0m\n",par[0]);   
         return;
   }
   if ((is_user(par)) && (strcasecmp(dcc[idx].nick, par) != 0)) {
      dprintf(idx,"[1;34mSomebody is already using %s.[0m\n",par);
      return;
   }
   if ((strcasecmp(par, origbotname) == 0) || (strcasecmp(par, botnetnick) == 0)) {
     dprintf(idx,"[1;34mHey!  That MY name![0m\n",par); 
         return;
   }
   icon = geticon(idx);
   if (change_handle(dcc[idx].nick, par)) {
      notes_change(idx, dcc[idx].nick, par);
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m nick %s",dcc[idx].nick,par);
      dprintf(idx,"[1;34mOkay, changed.[0m\n");
      if (dcc[idx].u.chat->channel >= 0) {
	 chanout2(dcc[idx].u.chat->channel, "Nick change: %s -> %s\n",
		  dcc[idx].nick, par);
	 context;
	 if (dcc[idx].u.chat->channel < 100000) {
	    tandout("part %s %s %d\n", botnetnick, dcc[idx].nick, dcc[idx].sock);
	    tandout("join %s %s %d %c%d %s\n", botnetnick, par,
	    dcc[idx].u.chat->channel, icon, dcc[idx].sock, dcc[idx].host);
	 }
      }
      for (i = 0; i < dcc_total; i++)
	 if ((idx != i) && (strcasecmp(dcc[idx].nick, dcc[i].nick) == 0))
	    strcpy(dcc[i].nick, par);
      strcpy(dcc[idx].nick, par);
   } else
   dprintf(idx,"[1;34mFailed.[0m\n");   
}

void cmd_pls_telnethost PROTO2(int, idx, char *, par)
 {
    char host[512];
    if (!par[0]) {
       dprintf(idx, "Usage: +telnethost <newhostmask>\n");
       return;
    }
    nsplit(host, par);
    sprintf(tmp3, "%s\n", host);
    tmp4 = 0;
    file=fopen(hostfile, "rt");
    sprintf(tmp, "%s~new", hostfile);
    file2=fopen(tmp, "wt");
    while (!feof(file)) {
    fgets(tmp2, 255, file);
    if (feof(file)) { break; }
    if (strcasecmp(tmp2, tmp3) == 0) { 
    tmp4 = 1; 
    }
    fputs(tmp2, file2);
    }
    fclose(file);
 
    if ( tmp4 == 1 ) {
    dprintf(idx, "The host '%s' is already in the hosts allowed to telnet in.\n", host);
    fclose(file2);
    }
 
    if ( tmp4 == 0 ) {
    putlog(LOG_CMDS, "*", "#%s# +telnethost %s ", dcc[idx].nick, host);
    dprintf(idx, "Added '%s' to allowed hosts to telnet in.\n", host);
    fputs(tmp3, file2);
    fclose(file2);
    movefile(tmp, hostfile);
    }
 }
 
 void cmd_mns_telnethost PROTO2(int, idx, char *, par)
 {
    char host[512];
    if (!par[0]) {
       dprintf(idx, "Usage: -telnethost <hostmask>\n");
       return;
    }
    nsplit(host, par);
    sprintf(tmp3, "%s\n", host);
    tmp4 = 0;
    file=fopen(hostfile, "rt");
    sprintf(tmp, "%s~new", hostfile);
    file2=fopen(tmp, "wt");
    while (!feof(file)) {
    fgets(tmp2, 255, file);
    if (feof(file)) { break; }
    if (strcasecmp(tmp2, tmp3) == 0) { 
    tmp4 = 1; 
    }
    if ( tmp4 == 0 ) { fputs(tmp2, file2); }
    if ( tmp4 == 2 ) { fputs(tmp2, file2); }
    if ( tmp4 == 1 ) { tmp4 = 2; }
    }
    fclose(file);
    fclose(file2);
 
    if ( tmp4 == 0 ) {
    dprintf(idx, "The host '%s' doesn't exist in the hosts allowed to telnet in.\n", host);
    }
 
    if ( tmp4 == 2 ) {
    putlog(LOG_CMDS, "*", "#%s# -telnethost %s ", dcc[idx].nick, host);
    dprintf(idx, "Removed '%s' from allowed hosts to telnet in.\n", host);
    movefile(tmp, hostfile);
    }
 }
 
void cmd_show_telnethost PROTO2(int, idx, char *, par)
 {
    file=fopen(hostfile, "rt");
    putlog(LOG_CMDS, "*", "#%s# telnethosts", dcc[idx].nick);
    dprintf(idx, "Hosts allowed to telnet in:\n");
    while (!feof(file)) {
    fgets(tmp2, 255, file);
    if (feof(file)) { break; }
    dprintf(idx, "%s", tmp2);
    }
    fclose(file);
}
void cmd_pls_host PROTO2(int, idx, char *, par)
{
   char handle[512], host[512];
   int chatr;
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: +host <handle> <newhostmask>[0m\n"); 
      return;
   }
   nsplit(handle, par);
   nsplit(host, par);
   chatr = get_chanattr_handle(handle, dcc[idx].u.chat->con_chan);
   if (!is_user(handle)) {
      dprintf(idx,"[1;34mNo such user.[0m\n");
      return;
   }
   if (ishost_for_handle(handle, host)) {
     dprintf(idx,"[1;34mThat hostmask is already there.[0m\n");
      return;
   }
   if ((get_attr_handle(dcc[idx].nick) & USER_BOTMAST) &&
       (!(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_OWNER))) &&
       (!(get_attr_handle(handle) & USER_BOT))) {
      dprintf(idx,"[1;34mYou can't add hostmasks to non-bots.[0m\n");
      return;
   }
   if (chatr == 0 && (!(get_attr_handle(dcc[idx].nick) & USER_MASTER))) {
    dprintf(idx,"[1;34mYou can't add hostmasks to non-channel users.[0m\n");  
         return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & USER_OWNER) &&
       !(get_attr_handle(dcc[idx].nick) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;34mCan't add hostmasks to the bot owner.[0m\n");
      return;
   }
#endif
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m +host %s %s",dcc[idx].nick,handle,host);
  addhost_by_handle(handle,host);
  dprintf(idx,"[1;34mAdded '%s' to %s[0m\n",host,handle);
}

void cmd_mns_host PROTO2(int, idx, char *, par)
{
   char handle[512], host[512];
   int chatr;
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: -host <handle> <hostmask>[0m\n");   
         return;
   }
   nsplit(handle, par);
   nsplit(host, par);
   chatr = get_chanattr_handle(handle, dcc[idx].u.chat->con_chan);
   if (!is_user(handle)) {
      dprintf(idx,"[1;34mNo such user.[0m\n");
      return;
   }
   if ((get_attr_handle(dcc[idx].nick) & USER_BOTMAST) &&
       (!(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_OWNER))) &&
       (!(get_attr_handle(handle) & USER_BOT))) {
     dprintf(idx,"[1;34mYou can't remove hostmasks from non-bots.[0m\n");
      return;
   }
   if (chatr == 0 && (!(get_attr_handle(dcc[idx].nick) & USER_MASTER))) {
    dprintf(idx,"[1;34mYou can't remove hostmasks from non-channel users.[0m\n");  
         return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & BOT_SHARE) &&
       (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
      dprintf(idx,"[1;34mYou can't remove hostmask from a shared bot.[0m\n");
      return;
   }
   if ((get_attr_handle(handle) & USER_OWNER) &&
       !(get_attr_handle(dcc[idx].nick) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;34mCan't remove hostmasks from the bot owner.[0m\n");
      return;
   }
#endif
   if (delhost_by_handle(handle, host)) {
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m -host %s %s",dcc[idx].nick,handle,host);
    dprintf(idx,"[1;34mRemoved '%s' from %s[0m\n",host,handle);  
  } else
  dprintf(idx,"[1;34mFailed.[0m\n");    
}

void cmd_chpass PROTO2(int, idx, char *, par)
{
   char handle[512], new[512];
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: chpass <handle> [password][0m\n");  
         return;
   }
   nsplit(handle, par);
   if (!par[0]) {
      if (!is_user(handle)) {
      dprintf(idx,"[1;34mNo such user.[0m\n");	 
      	 return;
      }
      if ((get_attr_handle(dcc[idx].nick) & USER_BOTMAST) &&
      (!(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_OWNER))) &&
	  (!(get_attr_handle(handle) & USER_BOT))) {
	 dprintf(idx,"[1;34mYou can't change passwords for non-bots.[0m\n");
	 return;
      }
#ifdef OWNER
      if ((get_attr_handle(handle) & BOT_SHARE) &&
	  (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
	 dprintf(idx,"[1;34mYou can't change shared bot's password.[0m\n");
	 return;
      }
      if ((get_attr_handle(handle) & USER_OWNER) &&
	  (strcasecmp(handle, dcc[idx].nick) != 0)) {
	 dprintf(idx,"[1;34mCan't change the bot owner's password.[0m\n");
	 return;
      }
#endif
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chpass %s [nothing]",dcc[idx].nick,handle);
      change_pass_by_handle(handle, "-");
      dprintf(idx,"[1;34mRemoved password.[0m\n");
      return;
}
   if (!is_user(handle)) {
   dprintf(idx,"[1;34mNo such user.[0m\n");   
         return;
   }
   if ((get_attr_handle(dcc[idx].nick) & USER_BOTMAST) &&
       (!(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_OWNER))) &&
       (!(get_attr_handle(handle) & USER_BOT))) {
      dprintf(idx,"[1;34mYou can't change password for non-bots.[0m\n");
      return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & BOT_SHARE) &&
       (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
      dprintf(idx,"[1;34mYou can't change shared bot's password.[0m\n");
      return;
   }
   if ((get_attr_handle(handle) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;34mCan't change the bot owner's password.[0m\n");
      return;
   }
#endif
   nsplit(new, par);
   if (strlen(new) > 16)
      new[16] = 0;
   if (strlen(new) < 4) {
   dprintf(idx,"[1;34mPlease use at least 4 characters.[0m\n");   
         return;
   }
   change_pass_by_handle(handle, new);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chpass %s [something]",dcc[idx].nick,handle);
   dprintf(idx,"[1;34mChanged password.[0m\n");
}

void cmd_chaddr PROTO2(int, idx, char *, par)
{
   char handle[512], addr[512];
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage: chaddr <botname> <host[:port#]>[0m\n");   
         return;
   }
   nsplit(handle, par);
   nsplit(addr, par);
   if (!(get_attr_handle(handle) & USER_BOT)) {
      dprintf(idx,"[1;34mUseful only for tandem bots.[0m\n");
      return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & BOT_SHARE) &&
       (!(get_attr_handle(dcc[idx].nick) & USER_OWNER))) {
      dprintf(idx,"[1;34mYou cant't change shared bot's address.[0m\n");
      return;
   }
#endif
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chaddr %s %s",dcc[idx].nick,handle,addr);
   dprintf(idx,"[1;34mChanged bot's address.[0m\n");
   set_handle_info(userlist, handle, addr);
}

void cmd_comment PROTO2(int, idx, char *, par)
{
   char handle[512];
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: comment <handle> <newcomment>[0m\n");  
         return;
   }
   nsplit(handle, par);
   if (!is_user(handle)) {
     dprintf(idx,"[1;34mNo such user![0m\n"); 
         return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;34mCan't change comment on the bot owner.[0m\n");
 return;
   }
#endif
 putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m comment %s %s",dcc[idx].nick,handle,par); 
   if (strcasecmp(par, "none") == 0) {
    dprintf(idx,"[1;34mOkay, comment blanked.[0m\n");  
         set_handle_comment(userlist, handle, "");
      return;
   }
   dprintf(idx,"[1;34mChanged comment.[0m\n");
   set_handle_comment(userlist, handle, par);
}

void cmd_url PROTO2(int, idx, char *, par)
{
   char s[161];
   if (!par[0]) {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m url",dcc[idx].nick);
      get_handle_url(dcc[idx].nick, s);
      if (s[0]) {
	 dprintf(idx,"[1;32mYour url address is: [1;33m%s[0m\n",s);
         dprintf(idx,"[1;31m(You can remove it with '.url none')[0m\n");
 } else
	 dprintf(idx,"[1;31mYou have no url address set.[0m\n");
 return;
   }
   if (strcasecmp(par, "none") == 0) {
   dprintf(idx,"[1;31mRemoved your url address.[0m\n");   
         set_handle_url(userlist, dcc[idx].nick, "");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m url %s",dcc[idx].nick,par);
   dprintf(idx,"[1;37mYour url address:[0m [1;50m%s[0m\n",par);
   set_handle_url(userlist, dcc[idx].nick, par);
}

void cmd_email PROTO2(int, idx, char *, par)
{
   char s[161];
   if (!par[0]) {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m email",dcc[idx].nick);
      get_handle_email(dcc[idx].nick, s);
      if (s[0]) {
	 dprintf(idx,"[1;31mYour email address is:[0m [1;33m%s[0m\n",s);
         dprintf(idx,"[1;31m(You can remove it with '.email none')[0m\n");
 } else
	 dprintf(idx,"[1;32mYou have no email address set.[0m\n");
 return;
   }
   if (strcasecmp(par, "none") == 0) {
   dprintf(idx,"[1;33mRemoved your email address.[0m\n");   
         set_handle_email(userlist, dcc[idx].nick, "");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m email %s",dcc[idx].nick,par);
   dprintf(idx,"[1;31mYour email address:[0m [1;50m%s[0m\n",par);
   set_handle_email(userlist, dcc[idx].nick, par);
}

void cmd_churl PROTO2(int, idx, char *, par)
{
   char handle[512];
   if (!par[0]) {
     dprintf(idx,"[1;32mUsage: churl <handle> [address][0m\n"); 
      return;
   }
   nsplit(handle, par);
   if (!is_user(handle)) {
     dprintf(idx,"[1;31mNo such user.[0m\n");
      return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;34mCan't change url address of the bot owner you doof.[0m\n");
      return;
   }
#endif
   if (!par[0]) {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m churl %s",dcc[idx].nick,handle);
      dprintf(idx,"[1;31mWiped url for[0m [1;50m%s[0m\n",handle);
      set_handle_url(userlist, handle, "");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m churl %s %s",dcc[idx].nick,handle,par);
   dprintf(idx,"[1;32mChanged url for[0m [1;33m %s[0m [1;37m to: %s[0m\n",handle,par);
   set_handle_url(userlist, handle, par);
}

void cmd_chemail PROTO2(int, idx, char *, par)
{
   char handle[512];
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage: chemail <handle> [address][0m\n"); 
      return;
   }
   nsplit(handle, par);
   if (!is_user(handle)) {
     dprintf(idx,"[1;34mNo such user.[0m\n");
      return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;34mCan't change email address of the bot owner.[0m\n");
      return;
   }
#endif
   if (!par[0]) {
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chemail %s",dcc[idx].nick,handle);
      dprintf(idx,"[1;34mWiped email for %s[0m\n",handle);
      set_handle_email(userlist, handle, "");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chemail %s %s",dcc[idx].nick,handle,par);
   dprintf(idx,"[1;34mChanged email for %s to: %s[0m\n",handle,par);
   set_handle_email(userlist, handle, par);
}

#ifndef NO_IRC
void cmd_dump PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
    dprintf(idx,"[1;34mUsage: dump <server stuff>[0m\n");      
         return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m dump %s",dcc[idx].nick,par);
   mprintf(serv, "%s\n", par);
}
#endif

void cmd_reset PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   int atr, chatr = 0;
   atr = get_attr_handle(dcc[idx].nick);
   if (par[0]) {
      chan = findchan(par);
      if (chan == NULL) {
	 dprintf(idx,"[1;34mI don't monitor channel %s[0m\n",par);
         return;
      }
      chatr = get_chanattr_handle(dcc[idx].nick, chan->name);
      if ((!(atr & USER_MASTER)) && (!(chatr & CHANUSER_MASTER))) {
	 dprintf(idx,"[1;34mYou are not a master on %s.[0m\n",chan->name);
         return;
      }
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m reset %s",dcc[idx].nick,par);
      dprintf(idx,"[1;34mResetting channel info for %s...[0m\n",par);
      reset_chan_info(chan);
      return;
   }
   if (!(atr & USER_MASTER)) {
      dprintf(idx,"[1;34mYou are not a Bot Master.[0m\n");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m reset all",dcc[idx].nick);
   dprintf(idx,"[1;34mResetting channel info for all channels...[0m\n");
   chan = chanset;
   while (chan != NULL) {
      reset_chan_info(chan);
      chan = chan->next;
   }
}

void cmd_restart PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m restart",dcc[idx].nick);
   if (!backgrd) {
     dprintf(idx,"[1;34mYou can not .restart a bot when running -n (due to tcl).[0m\n"); 
      return;
   }
   dprintf(idx,"[1;34mRestarting.[0m\n");
       if (make_userfile) {
      putlog(LOG_MISC, "*", "[1;34mUh, guess you don't need to create a new userfile.[0m\n");
      make_userfile = 0;
   }
   write_userfile();
   putlog(LOG_MISC, "*", "[1;34mRestarting....[0m\n");
   wipe_timers(interp, &utimer);
   wipe_timers(interp, &timer);
   do_restart = 1;
}

void cmd_rehash PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m rehash",dcc[idx].nick);
   dprintf(idx,"[1;34mRehashing.[0m\n");
   if (make_userfile) {
      putlog(LOG_MISC, "*", "Uh, guess you don't need to create a new userfile.");
      make_userfile = 0;
   }
   write_userfile();
   putlog(LOG_MISC, "*", "[1;34mRehashing....[0m\n");
   rehash();
}

void cmd_reload PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m reload",dcc[idx].nick);
   dprintf(idx,"[1;34mReloading user file...[0m\n");
   reload();
}

void cmd_die PROTO2(int, idx, char *, par)
{
   char s[512];
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m die %s",dcc[idx].nick,par);
  if (par[0]) {
    chatout("[1;31m*** BOT SHUTDOWN[0m ([1;35m%s: %s[0m)\n",dcc[idx].nick,par);
         tandout("chat %s BOT SHUTDOWN (%s: %s)\n", botnetnick, dcc[idx].nick, par);
      tprintf(serv, "QUIT :%s\n", par);
   } else {
      chatout("[1;31m*** BOT SHUTDOWN[0m ([1;35mauthorized by %s[0m)\n",dcc[idx].nick);
      tandout("chat %s BOT SHUTDOWN (authorized by %s)\n", botnetnick,
	      dcc[idx].nick);
      tprintf(serv, "QUIT :%s\n", dcc[idx].nick);
   }
   tandout("bye\n");
   write_userfile();
   sleep(3);			/* give the server time to understand */
   sprintf(s,"[1;31mDIE BY %s!%s[0m ([1;35m%s[0m)",dcc[idx].nick,dcc[idx].host,par[0]?par:
   	   "request");
   fatal(s, 0);
}

#ifndef NO_IRC
void cmd_jump PROTO2(int, idx, char *, par)
{
   char other[512], port[512];
   if (par[0]) {
      nsplit(other, par);
      nsplit(port, par);
      if (!port[0])
	 sprintf(port, "%d", default_port);
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m jump %s %s %s",dcc[idx].nick,other,port,par);
      strcpy(newserver, other);
      newserverport = atoi(port);
      strcpy(newserverpass, par);
   } else
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m jump",dcc[idx].nick);   
   dprintf(idx,"[1;34mJumping servers...[0m\n");
   tprintf(serv,"QUIT :brb\n");
   if (serv >= 0)
      killsock(serv);
   serv = (-1);
}
#endif

void cmd_debug PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m debug",dcc[idx].nick);
   debug_mem_to_dcc(idx);
}

void cmd_info PROTO2(int, idx, char *, par)
{
   char s[512], chname[512];
   int locked = 0;
   if (!use_info) {
      dprintf(idx,"[1;34mInfo storage is turned off.[0m\n");
      return;
   }
   /* yes, if the default info line is locked, then all the channel ones */
   /* are too. */
   get_handle_info(dcc[idx].nick, s);
   if (s[0] == '@')
      locked = 1;
   if ((par[0] == '#') || (par[0] == '+') || (par[0] == '&')) {
      nsplit(chname, par);
      if (!defined_channel(chname)) {
	 dprintf(idx,"[1;34mNo such channel.[0m\n");
         return;
      }
      get_handle_chaninfo(dcc[idx].nick, chname, s);
      if (s[0] == '@')
	 locked = 1;
   } else
      chname[0] = 0;
   if (!par[0]) {
      if (s[0] == '@')
	 strcpy(s, &s[1]);
      if (s[0]) {
	 if (chname[0]) {
	    dprintf(idx,"[1;34mInfo on %s[0m: [1;31m%s[0m\n",chname,s);
            dprintf(idx,"[1;34mUse[0m '[1;31m.info %s none[0m' to remove it.[0m\n",chname);
	     } else {
	    dprintf(idx,"[1;34mDefault info[0m: [1;31m%s[0m\n",s);
            dprintf(idx,"[1;34mUse[0m '[1;31m.info none[0m' [1;34mto remove it.[0m\n");
 }
      } else
     dprintf(idx,"[1;33mNo info has been set for you.[0m\n");
     putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m info %s",dcc[idx].nick,chname);	 
      return;
   }
   if ((locked) && !(get_attr_handle(dcc[idx].nick) & USER_MASTER)) {
      dprintf(idx,"[1;31mYour info line is locked.  Sorry.[0m\n");
      return;
   }
   if (strcasecmp(par, "none") == 0) {
      par[0] = 0;
      if (chname[0]) {
	 set_handle_chaninfo(userlist, dcc[idx].nick, chname, par);
	 dprintf(idx,"[1;32mRemoved your info line on %s.[0m\n",chname);
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m info %s none",dcc[idx].nick,chname);
 } else {
	 set_handle_info(userlist, dcc[idx].nick, par);
	 dprintf(idx,"[1;33mRemoved your default info line.[0m\n");
         putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m info none",dcc[idx].nick);
 }
      return;
   }
   if (par[0] == '@')
      strcpy(par, &par[1]);
   if (chname[0]) {
      set_handle_chaninfo(userlist, dcc[idx].nick, chname, par);
      dprintf(idx,"[1;34mYour info on %s is now[0m: [1;31m%s[0m\n",chname,par);
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m info %s ...",dcc[idx].nick,chname);
 } else {
      set_handle_info(userlist, dcc[idx].nick, par);
     dprintf(idx,"[1;34mYour default info is now[0m: [1;31m%s[0m\n",par);
    putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m info ...",dcc[idx].nick);
 }
}

void cmd_chinfo PROTO2(int, idx, char *, par)
{
   char handle[512], chname[512];
   if (!use_info) {
      dprintf(idx,"[1;34mInfo storage is turned off.[0m\n");
      return;
   }
   nsplit(handle, par);
   if (!handle[0]) {
      dprintf(idx,"[1;34mUsage: chinfo <handle> [channel] <new-info>[0m\n");
      return;
   }
   if (!is_user(handle)) {
      dprintf(idx,"[1;34mNo such user.[0m\n");
      return;
   }
   if ((par[0] == '#') || (par[0] == '+') || (par[0] == '&'))
      nsplit(chname, par);
   else
      chname[0] = 0;
   if (get_attr_handle(handle) & USER_BOT) {
      dprintf(idx,"[1;31mUseful only for users.[0m\n");
      return;
   }
#ifdef OWNER
   if ((get_attr_handle(handle) & USER_OWNER) &&
       (strcasecmp(handle, dcc[idx].nick) != 0)) {
      dprintf(idx,"[1;31mCan't change info for the bot owner.[0m\n");
      return;
   }
#endif
   if (chname[0])
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chinfo %s %s %s",dcc[idx].nick,handle,chname,par);
 else
     putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chinfo %s %s",dcc[idx].nick,handle,par); 
      if (strcasecmp(par, "none") == 0)
      par[0] = 0;
   if (chname[0]) {
      set_handle_chaninfo(userlist, handle, chname, par);
      if (par[0] == '@')
	 dprintf(idx,"[1;33mNew info[0m ([1;50mLOCKED[0m)[1;34m for %s on %s[0m: [1;31m%s[0m\n",handle,chname,&par[1]);
          else if (par[0])
	 dprintf(idx,"[1;33mNew info for %s on %s[0m: [1;31m%s[0m\n",handle,chname,par);
    else
        dprintf(idx,"[1;33mWiped info for %s on %s[0m\n",handle,chname);	 
 } else {
      set_handle_info(userlist, handle, par);
      if (par[0] == '@')
	 dprintf(idx,"[1;33mNew default info[0m ([1;50mLOCKED[0m) [1;34mfor %s[0m: [1;31m%s[0m\n",handle,&par[1]);
     else if (par[0])
       dprintf(idx,"N[1;33mNew default info for[0m [1;50m%s[0m: [1;31m%s[0m\n",handle,par);	 
        else
	dprintf(idx,"[1;33mWiped default info for[0m [1;50m%s[0m\n",handle); 
    }
}

#ifdef ENABLE_SIMUL
void cmd_simul PROTO2(int, idx, char *, par)
{
   char nick[512];
   int i, ok = 0;
   nsplit(nick, par);
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage:[0m [1;50msimul <nick> <test>[0m\n");
      return;
   }
   for (i = 0; i < dcc_total; i++)
      if ((strcasecmp(nick, dcc[i].nick) == 0) && (!ok) &&
	  ((dcc[i].type == DCC_CHAT) || (dcc[i].type == DCC_FILES))) {
	 putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m simul %s %s",dcc[idx].nick,nick,par);
	 dcc_activity(dcc[i].sock, par, strlen(par));
	 ok = 1;
      }
   if (!ok)
      dprintf(idx,"[1;33mNo such user on the party line.[0m\n");
}
#endif

void cmd_link PROTO2(int, idx, char *, par)
{
   char s[512];
   int i;
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage:[0m [1;50mlink [some-bot] <new-bot>[0m\n");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m link %s",dcc[idx].nick,par);
   if (strchr(par, ' ') == NULL)
      botlink(dcc[idx].nick, idx, par);
   else {
      split(s, par);
      i = nextbot(s);
      if (i < 0) {
      dprintf(idx,"[1;32mNo such bot online.[0m\n");	 
      	 return;
      }
      tprintf(dcc[i].sock, "link %d:%s@%s %s %s\n", dcc[idx].sock, dcc[idx].nick,
	      botnetnick, s, par);
   }
}

void cmd_unlink PROTO2(int, idx, char *, par)
{
   int i;
   char bot[200];
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage: unlink <bot> [reason][0m\n");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m unlink %s",dcc[idx].nick,par);
   nsplit(bot, par);
   i = nextbot(bot);
   if (i < 0) {
      botunlink(idx, bot, par);
      return;
   }
   /* if we're directly connected to that bot, just do it */
   if (strcasecmp(dcc[i].nick, bot) == 0)
      botunlink(idx, bot, par);
   else {
      tprintf(dcc[i].sock, "unlink %d:%s@%s %s %s %s\n", dcc[idx].sock,
	      dcc[idx].nick, botnetnick, lastbot(bot), bot, par);
   }
}

void cmd_relay PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
     dprintf(idx,"[1;34mUsage:[0m [1;50mrelay <bot>[0m\n"); 
         return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m relay %s",dcc[idx].nick,par);
   tandem_relay(idx, par);
}

void cmd_save PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m save",dcc[idx].nick);
   write_userfile();
}

void cmd_trace PROTO2(int, idx, char *, par)
{
   int i;
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage:[0m [1;50mtrace <botname>[0m\n");
      return;
   }
   if (strcasecmp(par, botnetnick) == 0) {
      dprintf(idx,"[1;33mThat's me, doof! (=[0m\n");
      return;
   }
   i = nextbot(par);
   if (i < 0) {
      dprintf(idx,"[1;32mUnreachable bot.[0m\n");
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m trace %s",dcc[idx].nick,par);
   tprintf(dcc[i].sock, "trace %d:%s@%s %s %s\n", dcc[idx].sock, dcc[idx].nick,
	   botnetnick, par, botnetnick);
}

#ifndef NO_IRC
void cmd_topic PROTO2(int, idx, char *, par)
{
   struct chanset_t *chan;
   if (!has_op(idx, ""))
      return;
   chan = findchan(dcc[idx].u.chat->con_chan);
   if (!par[0]) {
      if (chan->channel.topic[0] == 0) {
	 dprintf(idx,"[1;33mNo topic is set for %s[0m\n", chan->name);
         return;
      } else {
	dprintf(idx,"[1;34mThe topic for[0m [1;50m%s is:[0m [1;37m%s[0m\n", chan->name, chan->channel.topic); 
     return;
      }
   }
   if ((channel_optopic(chan)) && (!me_op(chan))) {
      dprintf(idx,"[1;31mI'm not a channel op on[0m [1;32m%s[0m [1;31mand the channel is +t.[0m\n",
      	      chan->name);
      return;
   }
   mprintf(serv, "TOPIC %s :%s\n", chan->name, par);
   dprintf(idx,"[1;33mChanging topic...[0m\n");
   strcpy(chan->channel.topic, par);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m (%s) topic %s",dcc[idx].nick,
   	  dcc[idx].u.chat->con_chan, par);
}
#endif

void cmd_binds PROTO2(int, idx, char *, par)
{
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m binds %s",dcc[idx].nick,par);
   tell_binds(idx, par);
}

void cmd_banner PROTO2(int, idx, char *, par)
{
   char s[540];
   int i;
   if (!par[0]) {
   dprintf(idx,"[1;34mUsage:[0m [1;50mbanner <message>[0m\n");   
         return;
   }
   sprintf(s, "\007\007### Botwide:[%s] %s\n", dcc[idx].nick, par);
   for (i = 0; i < dcc_total; i++)
      if ((dcc[i].type == DCC_CHAT) || (dcc[i].type == DCC_FILES))
	 dprintf(i, s);
}

/* after messing with someone's user flags, make sure the dcc-chat flags
   are set correctly */
int check_dcc_attrs(char *hand, int flags, int oatr)
{
   int i, stat;
#ifdef OWNER
/* if it matches someone in the owner list, make sure he/she has +n */
   if (owner[0]) {
      char *p, s[512];
      strcpy(s, owner);
      p = strchr(s, ',');
      while (p != NULL) {
	 *p = 0;
	 rmspace(s);
	 if (strcasecmp(hand, s) == 0)
	    flags |= USER_OWNER;
	 strcpy(s, p + 1);
	 p = strchr(s, ',');
      }
      rmspace(s);
      if (strcasecmp(hand, s) == 0)
	 flags |= USER_OWNER;
      set_attr_handle(hand, flags);
   }
#endif
   for (i = 0; i < dcc_total; i++) {
      if (((dcc[i].type == DCC_CHAT) || (dcc[i].type == DCC_FILES)) &&
	  (strcasecmp(hand, dcc[i].nick) == 0)) {
	 if (dcc[i].type == DCC_CHAT)
	    stat = dcc[i].u.chat->status;
	 else
	    stat = dcc[i].u.file->chat->status;
	 if ((dcc[i].type == DCC_CHAT) && ((flags &
					    (USER_GLOBAL | USER_MASTER | USER_OWNER | USER_BOTMAST)) != (oatr &
	     (USER_GLOBAL | USER_MASTER | USER_OWNER | USER_BOTMAST)))) {
	    tandout("part %s %s %d\n", botnetnick, dcc[i].nick, dcc[i].sock);
	    tandout("join %s %s %d %c%d %s\n", botnetnick, dcc[i].nick,
	    dcc[i].u.chat->channel, geticon(i), dcc[i].sock, dcc[i].host);
	 }
	 if ((oatr & USER_MASTER) && !(flags & USER_MASTER)) {
	    dcc[i].u.chat->con_flags &= ~(LOG_MISC | LOG_CMDS | LOG_RAW | LOG_FILES |
			      LOG_LEV1 | LOG_LEV2 | LOG_LEV3 | LOG_LEV4 |
			      LOG_LEV5 | LOG_LEV6 | LOG_LEV7 | LOG_LEV8 |
					  LOG_WALL | LOG_DEBUG);
	    if (master_anywhere(hand))
	       dcc[i].u.chat->con_flags |= (LOG_MISC | LOG_CMDS);
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are no longer a master on this bot.[0m\n");
 }
	 if (!(oatr & USER_MASTER) && (flags & USER_MASTER)) {
	    dcc[i].u.chat->con_flags |= conmask;
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are now a master on this bot.\n[m0");
 }
	 if (!(oatr & USER_BOTMAST) && (flags & USER_BOTMAST)) {
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are now a botnet master on this bot.\n[m0");
 }
	 if ((oatr & USER_BOTMAST) && !(flags & USER_BOTMAST)) {
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are no longer a botnet master on this bot.[0m\n");
 }
#ifdef OWNER
	 if (!(oatr & USER_OWNER) && (flags & USER_OWNER)) {
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are now an OWNER of this bot.\n[m0");
 }
	 if ((oatr & USER_OWNER) && !(flags & USER_OWNER)) {
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are no longer an owner of this bot.\n[m0");
 }
#endif
	 if ((stat & STAT_PARTY) && (flags & USER_GLOBAL))
	    stat &= ~STAT_PARTY;
	 if (!(stat & STAT_PARTY) && !(flags & USER_GLOBAL) && !(flags & USER_MASTER))
	    stat |= STAT_PARTY;
	 if ((stat & STAT_CHAT) && !(flags & USER_PARTY) && !(flags & USER_MASTER) &&
	     (!(flags & USER_GLOBAL) || require_p))
	    stat &= ~STAT_CHAT;
	 if ((dcc[i].type == DCC_FILES) && !(stat & STAT_CHAT) &&
	     ((flags & USER_MASTER) || (flags & USER_PARTY) ||
	      ((flags & USER_GLOBAL) && !require_p)))
	    stat |= STAT_CHAT;
	 if (dcc[i].type == DCC_CHAT)
	    dcc[i].u.chat->status = stat;
	 else
	    dcc[i].u.file->chat->status = stat;
	 /* check if they no longer have access to wherever they are */
	 /* DON'T kick someone off the party line just cos they lost +p */
	 /* (pinvite script removes +p after 5 mins automatically) */
	 if ((dcc[i].type == DCC_FILES) && !(flags & USER_XFER) &&
	     !(flags & USER_MASTER)) {
	    dprintf(i,"[1;31m-+- POOF! -+-[0m\n");
	    dprintf(i,"[1;33mYou no longer have file area access.[0m\n\n");
	    putlog(LOG_MISC, "*", "DCC user [%s]%s removed from file system",
		   dcc[i].nick, dcc[i].host);
	    if (dcc[i].u.file->chat->status & STAT_CHAT) {
	       struct chat_info *ci;
	       ci = dcc[i].u.file->chat;
	       nfree(dcc[i].u.file);
	       dcc[i].u.chat = ci;
	       dcc[i].u.chat->status &= (~STAT_CHAT);
	       dcc[i].type = DCC_CHAT;
	       if (dcc[i].u.chat->channel >= 0) {
		  chanout2(dcc[i].u.chat->channel, "%s has returned.\n", dcc[i].nick);
		  context;
		  if (dcc[i].u.chat->channel < 100000)
		     tandout("unaway %s %d\n", botnetnick, dcc[i].sock);
	       }
	    } else {
	       killsock(dcc[i].sock);
	       dcc[i].sock = dcc[i].type;
	       dcc[i].type = DCC_LOST;
	       tandout("part %s %s %d\n", botnetnick, dcc[i].nick, dcc[i].sock);
	    }
	 }
      }
      if ((dcc[i].type == DCC_BOT) && (strcasecmp(hand, dcc[i].nick) == 0)) {
	 if ((dcc[i].u.bot->status & STAT_LEAF) && !(flags & BOT_LEAF))
	    dcc[i].u.bot->status &= ~(STAT_LEAF | STAT_WARNED);
	 if (!(dcc[i].u.bot->status & STAT_LEAF) && (flags & BOT_LEAF))
	    dcc[i].u.bot->status |= STAT_LEAF;
      }
   }
   return flags;
}

int check_dcc_chanattrs(char *hand, char *chname, int chflags, int ochatr)
{
   int i, chatr, found = 0;
   struct chanset_t *chan;
   chan = chanset;
   for (i = 0; i < dcc_total; i++) {
      if (((dcc[i].type == DCC_CHAT) || (dcc[i].type == DCC_FILES)) &&
	  (strcasecmp(hand, dcc[i].nick) == 0)) {
	 if ((dcc[i].type == DCC_CHAT) && ((chflags &
	   (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER)) != (ochatr &
		    (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER)))) {
	    tandout("part %s %s %d\n", botnetnick, dcc[i].nick, dcc[i].sock);
	    tandout("join %s %s %d %c%d %s\n", botnetnick, dcc[i].nick,
	    dcc[i].u.chat->channel, geticon(i), dcc[i].sock, dcc[i].host);
	 }
	 if ((ochatr & CHANUSER_MASTER) && !(chflags & CHANUSER_MASTER)) {
	    if (!(get_attr_handle(hand) & USER_MASTER))
	       dcc[i].u.chat->con_flags &= ~(LOG_MISC | LOG_CMDS);
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are no longer a master on %s.\n[m0", chname);
	    	 }
	 if (!(ochatr & CHANUSER_MASTER) && (chflags & CHANUSER_MASTER)) {
	    dcc[i].u.chat->con_flags |= conmask;
	    if (!(get_attr_handle(hand) & USER_MASTER))
	       dcc[i].u.chat->con_flags &= ~(LOG_LEV1 | LOG_LEV2 | LOG_LEV3 | LOG_LEV4 |
			      LOG_LEV5 | LOG_LEV6 | LOG_LEV7 | LOG_LEV8 |
			     LOG_RAW | LOG_DEBUG | LOG_WALL | LOG_FILES);
	    dprintf(i,"[1;31m*** POOF! ***[0m\n");
	    dprintf(i,"[1;33mYou are now a master on %s.\n[0m", chname);
 }
#ifdef OWNER
	 if (!(ochatr & CHANUSER_OWNER) && (chflags & CHANUSER_OWNER)) {
	    dprintf(i,"[1;31m@@@ POOF! @@@[0m\n");
	    dprintf(i,"[1;33mYou are now an OWNER of %s.[0m\n", chname);
 }
	 if ((ochatr & CHANUSER_OWNER) && !(chflags & CHANUSER_OWNER)) {
	    dprintf(i,"[1;31m@@@ POOF! @@@[0m\n");
	    dprintf(i,"[1;33mYou are no longer an owner of %s.[0m\n", chname);
 }
#endif
	 if (((ochatr & (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER)) &&
	      (!(chflags & (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER)))) ||
	 ((chflags & (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER)) &&
	  (!(ochatr & (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER))))) {
	    while (chan != NULL && found == 0) {
	       chatr = get_chanattr_handle(dcc[i].nick, chan->name);
	       if (chatr & (CHANUSER_OP | CHANUSER_MASTER | CHANUSER_OWNER))
		  found = 1;
	       else
		  chan = chan->next;
	    }
	    if (chan == NULL)
	       chan = chanset;
	    strcpy(dcc[i].u.chat->con_chan, chan->name);
	 }
      }
   }
   return chflags;
}

/* some flags are mutually exclusive -- this roots them out */
int sanity_check(int atr)
{
   if ((atr & BOT_HUB) && (atr & BOT_ALT))
      atr &= ~BOT_ALT;
   if (atr & BOT_REJECT) {
      if (atr & BOT_SHARE)
	 atr &= ~(BOT_SHARE | BOT_REJECT);
      if (atr & BOT_HUB)
	 atr &= ~(BOT_HUB | BOT_REJECT);
      if (atr & BOT_ALT)
	 atr &= ~(BOT_ALT | BOT_REJECT);
   }
   if (atr & USER_BOT) {
      if (atr & (USER_PARTY | USER_MASTER | USER_COMMON | USER_OWNER))
	 atr &= ~(USER_PARTY | USER_MASTER | USER_COMMON | USER_OWNER);
   } else {
      if (atr & (BOT_LEAF | BOT_REJECT | BOT_ALT | BOT_SHARE | BOT_HUB))
	 atr &= ~(BOT_LEAF | BOT_REJECT | BOT_ALT | BOT_SHARE | BOT_HUB);
   }
   if ((atr & USER_GLOBAL) && (atr & USER_DEOP))
      atr &= ~(USER_GLOBAL | USER_DEOP);
#ifdef OWNER
   /* can't be owner without also being master */
   if (atr & USER_OWNER)
      atr |= USER_MASTER;
#endif
   /* can't be botnet master without party-line access */
   if (atr & USER_BOTMAST)
      atr |= USER_PARTY;
   return atr;
}

/* sanity check on channel attributes */
int chan_sanity_check(chatr)
int chatr;
{
   if ((chatr & CHANUSER_OP) && (chatr & CHANUSER_DEOP))
      chatr &= ~(CHANUSER_OP | CHANUSER_DEOP);
#ifdef OWNER
   /* can't be channel owner without also being channel master */
   if (chatr & CHANUSER_OWNER)
      chatr |= CHANUSER_MASTER;
#endif
   return chatr;
}

void cmd_chattr PROTO2(int, idx, char *, par)
{
   char hand[512], s[21], chg[512];
   struct chanset_t *chan;
   int i, pos, f, atr, oatr, chatr = 0, ochatr = 0, own, chown = 0;
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage:[0m [1;50mchattr <handle> [changes [channel]][0m\n");
      return;
   }
   nsplit(hand, par);
   if ((hand[0] == '*') || (!is_user(hand))) {
     dprintf(idx,"[1;37mNo such user![0m\n");
      return;
   }
   chg[0] = 0;
   own = (get_attr_handle(dcc[idx].nick) & USER_OWNER);
   oatr = atr = get_attr_handle(hand);
   pos = 1;
   if (par[0]) {
      /* make changes */
      nsplit(chg, par);
      if (par[0]) {
	 if ((!defined_channel(par)) && (par[0] != '*')) {
	    dprintf(idx,"[1;33mNo channel record for %s[0m\n",par);
	    return;
	 }
      }
      if (!par[0] && (!(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_BOTMAST)))
      && (!(get_chanattr_handle(dcc[idx].nick, par) & CHANUSER_MASTER))) {
	 dprintf(idx,"[1;32mYou do not have Bot Master privileges[0m\n");
         return;
      }
      if (!(get_chanattr_handle(dcc[idx].nick, par) & CHANUSER_MASTER) &&
      !(get_attr_handle(dcc[idx].nick) & (USER_MASTER | USER_BOTMAST))) {
	dprintf(idx,"[1;32mYou do not have channel master privileges for channel %s[0m\n", par); 
         return;
      }
      if (par[0]) {
	 if (par[0] == '*') {
	    chown = 0;
	    ochatr = chatr = get_chanattr_handle(hand, chanset->name);
	 } else {
	    chown = (get_chanattr_handle(dcc[idx].nick, par) & CHANUSER_OWNER);
	    ochatr = chatr = get_chanattr_handle(hand, par);
	 }
      }
      for (i = 0; i < strlen(chg); i++) {
	 if (chg[i] == '+')
	    pos = 1;
	 else if (chg[i] == '-')
	    pos = 0;
	 else {
	    s[1] = 0;
	    s[0] = chg[i];
	    if (par[0]) {
	       /* channel-specific */
	       f = str2chflags(s);
#ifdef OWNER
	       if (!own && !chown)
		  f &= ~(CHANUSER_MASTER | CHANUSER_OWNER);
	       /* only owners and channel owners can +/- channel masters and owners */
#endif
	       change_chanflags(userlist, hand, par, (pos ? f : 0), (pos ? 0 : f));
	       /* FIXME: other stuff here? */
	    } else {
	       f = str2flags(s);
	       /* nobody can modify the +b bot flag */
	       f &= ~USER_BOT;
	       /* nobody can modify the +M, +N, and +O psuedo flags */
	       f &= ~(USER_PSUEDOOP | USER_PSUMST | USER_PSUOWN);
#ifdef OWNER
	       if (!own)
		  f &= (~(USER_OWNER | USER_MASTER | USER_BOTMAST | BOT_SHARE));
	       /* only owner can do +/- owner,master,share,botnet master */
#endif
	       atr = pos ? (atr | f) : (atr & ~f);
	    }
	 }
      }
      if (!par[0]) {
	 noshare = 1;
	 atr = sanity_check(atr);
	 atr = check_dcc_attrs(hand, atr, oatr);
	 noshare = 0;
	 set_attr_handle(hand, atr);
      }
   }
   if (par[0])
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chattr %s %s %s",dcc[idx].nick,hand,chg,par);
 else if (chg[0])
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chattr %s %s",dcc[idx].nick,hand, chg);
 else
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chattr %s",dcc[idx].nick,hand);   
   /* get current flags and display them */
   if (!par[0]) {
      flags2str(atr, s);
      dprintf(idx,"[1;33mFlags for[0m [1;50m%s[0m [1;31mare now +%s[0m\n",hand,s);
      chan = chanset;
      while (chan != NULL) {
	 recheck_channel(chan);
	 chan = chan->next;
      }
   } else if (par[0] == '*') {
      /* every channel */
      chan = chanset;
      while (chan != NULL) {
	 chatr = get_chanattr_handle(hand, chan->name);
	 chatr = chan_sanity_check(chatr);
	 if (chan == chanset)
	    chatr = check_dcc_chanattrs(hand, chan->name, chatr, ochatr);
	 chflags2str(chatr, s);
	 dprintf(idx,"[1;31mFlags for[0m [1;50m%s[0m [1;37mare now +%s %s[0m\n",hand,s,chan->name);
	 set_chanattr_handle(hand, chan->name, chatr);
	 recheck_channel(chan);
	 chan = chan->next;
      }
   } else {
      chatr = get_chanattr_handle(hand, par);
      chatr = chan_sanity_check(chatr);
      chatr = check_dcc_chanattrs(hand, par, chatr, ochatr);
      chflags2str(chatr, s);
      dprintf(idx,"[1;31mFlags for[0m [1;50m%s[0m [1;37mare now +%s %s[0m\n",hand,s,par);
      set_chanattr_handle(hand, par, chatr);
      recheck_channel(findchan(par));
   }
}

void cmd_notes PROTO2(int, idx, char *, par)
{
   char fcn[512];
   if (!par[0]) {
      dprintf(idx,"[1;31mUsage: notes index[0m\n");
      dprintf(idx,"[1;37m       notes read <# or ALL>[0m\n");
      dprintf(idx,"[1;33m       notes erase <# or ALL>[0m\n");
      return;
   }
   nsplit(fcn, par);
   if (strcasecmp(fcn, "index") == 0)
      notes_read(dcc[idx].nick, "", -1, idx);
   else if (strcasecmp(fcn, "read") == 0) {
      if (strcasecmp(par, "all") == 0)
	 notes_read(dcc[idx].nick, "", 0, idx);
      else
	 notes_read(dcc[idx].nick, "", atoi(par), idx);
   } else if (strcasecmp(fcn, "erase") == 0) {
      if (strcasecmp(par, "all") == 0)
	 notes_del(dcc[idx].nick, "", 0, idx);
      else
	 notes_del(dcc[idx].nick, "", atoi(par), idx);
   } else
       dprintf(idx,"[1;33mFunction must be one of INDEX, READ, or ERASE.[0m\n");
       putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m notes %s %s",dcc[idx].nick,fcn,par);
}

void cmd_chat PROTO2(int, idx, char *, par)
{
   int newchan, oldchan;
   if (strcasecmp(par, "off") == 0) {
      /* turn chat off */
      if (dcc[idx].u.chat->channel < 0)
	 dprintf(idx,"[1;34mYou weren't in chat anyway![0m\n");
     else {
	 dprintf(idx,"[1;34mLeaving chat mode...[0m\n");
         check_tcl_chpt(botnetnick, dcc[idx].nick, dcc[idx].sock);
	 chanout2(dcc[idx].u.chat->channel, "%s left the party line.\n",
		  dcc[idx].nick);
	 context;
	 if (dcc[idx].u.chat->channel < 100000)
	    tandout("part %s %s %d\n", botnetnick, dcc[idx].nick, dcc[idx].sock);
      }
      dcc[idx].u.chat->channel = (-1);
   } else {
      if (par[0] == '*') {
	 if (((par[1] < '0') || (par[1] > '9'))) {
	    if (par[1] == 0)
	       newchan = 0;
	    else
	       newchan = get_assoc(par + 1);
	    if (newchan < 0) {
	       dprintf(idx,"[1;34mNo channel by that name.[0m\n");
	       return;
	    }
	 } else
	    newchan = 100000 + atoi(par + 1);
	 if (newchan < 100000 || newchan > 199999) {
	  dprintf(idx,"[1;34mChannel # out of range: must be between 0 and 99999.[0m\n");  
    return;
	 }
      } else {
	 if (((par[0] < '0') || (par[0] > '9')) && (par[0])) {
	    if (strcasecmp(par, "on") == 0)
	       newchan = 0;
	    else
	       newchan = get_assoc(par);
	    if (newchan < 0) {
	      dprintf(idx,"[1;34mNo channel by that name.[0m\n"); 
	       return;
	    }
	 } else
	    newchan = atoi(par);
	 if ((newchan < 0) || (newchan > 99999)) {
	    dprintf(idx,"[1;34mChannel # out of range: must be between 0 and 99999.[0m\n");
	    return;
	 }
      }
      /* if coming back from being off the party line, make sure they're
         not away */
      if ((dcc[idx].u.chat->channel < 0) && (dcc[idx].u.chat->away != NULL))
	 not_away(idx);
      if (dcc[idx].u.chat->channel == newchan) {
	 if (newchan == 0)
	   dprintf(idx,"[1;32mYou're already on the party line![0m\n"); 
	 else
	 dprintf(idx,"[1;33mYou're already on channel %s%d![0m\n",   
	 	    (newchan < 100000) ? "" : "*", newchan % 100000);
      } else {
	 oldchan = dcc[idx].u.chat->channel;
	 if (oldchan >= 0)
	    check_tcl_chpt(botnetnick, dcc[idx].nick, dcc[idx].sock);
	 if (oldchan == 0) {
	    chanout2(0, "%s left this lifeless realm.\n", dcc[idx].nick);
	    context;
	 } else if (oldchan > 0) {
	    chanout2(oldchan, "%s left this lame ass channel.\n", dcc[idx].nick);
	    context;
	 }
	 dcc[idx].u.chat->channel = newchan;
	 if (newchan == 0) {
	 dprintf(idx,"[1;31mEntering the party line...[0m\n");   
	     chanout2(0, "%s entered this lifeless realm!.\n", dcc[idx].nick);
	    context;
	 } else {
	    dprintf(idx,"[1;32mJoining channel '%s'...[0m\n",par);
	    chanout2(newchan, "%s got his ass on dis channel.\n", dcc[idx].nick);
	    context;
	 }
	 check_tcl_chjn(botnetnick, dcc[idx].nick, newchan, geticon(idx),
			dcc[idx].sock, dcc[idx].host);
	 if (newchan < 100000)
	    tandout("join %s %s %d %c%d %s\n", botnetnick, dcc[idx].nick, newchan,
		    geticon(idx), dcc[idx].sock, dcc[idx].host);
	 else if (oldchan < 100000)
	    tandout("part %s %s %d\n", botnetnick, dcc[idx].nick, dcc[idx].sock);
      }
   }
}

void cmd_echo PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
      dprintf(idx,"[1;33mEcho is currently %s.[0m\n",dcc[idx].u.chat->status&STAT_ECHO?
      	      "on" : "off");
      return;
   }
   if (strcasecmp(par, "on") == 0) {
      dprintf(idx,"[1;31mEcho turned on.[0m\n");
      dcc[idx].u.chat->status |= STAT_ECHO;
      return;
   }
   if (strcasecmp(par, "off") == 0) {
      dprintf(idx,"[1;37mEcho turned off.[0m\n");
      dcc[idx].u.chat->status &= ~STAT_ECHO;
      return;
   }
   dprintf(idx,"[1;34mUsage:[0m [1;50mecho <on/off>[0m\n");
}

void cmd_fries PROTO2(int, idx, char *, par)
{
   dprintf(idx,"[1;31m*[0m [1;32m%s juliennes some fries for you.[0m\n",botnetnick);
   dprintf(idx,"[1;33mEnjoy![0m\n");
}
  
void cmd_beer PROTO2(int, idx, char *, par)
{
   dprintf(idx,"[1;31m*[0m [1;32m%s throws you a cold brewsky.[0m\n",botnetnick);
   dprintf(idx,"[1;37mIt's The King of beers Bud Enjoy![0m\n");
}
   
void cmd_flush PROTO2(int, idx, char *, par)
{
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage:[0m [1;50mflush <botname>[0m\n");
      return;
   }
   if (flush_tbuf(par))
   dprintf(idx,"[1;34mFlushed resync buffer for %s[0m\n",par);
  else dprintf(idx,"[1;34mThere is no resync buffer for that bot.[0m\n");   
}

int stripmodes(char *s)
{
   int i;
   int res = 0;
   for (i = 0; i < strlen(s); i++)
      switch (tolower(s[i])) {
      case 'b':
	 res |= STRIP_BOLD;
	 break;
      case 'c':
	 res |= STRIP_COLOR;
	 break;
      case 'r':
	 res |= STRIP_REV;
	 break;
      case 'u':
	 res |= STRIP_UNDER;
	 break;
      case 'a':
	 res |= STRIP_ANSI;
	 break;
      case '*':
	 res |= STRIP_ALL;
	 break;
      }
   return res;
}

char *stripmasktype(int x)
{
   static char s[20];
   char *p = s;
   if (x & STRIP_BOLD)
      *p++ = 'b';
   if (x & STRIP_COLOR)
      *p++ = 'c';
   if (x & STRIP_REV)
      *p++ = 'r';
   if (x & STRIP_UNDER)
      *p++ = 'u';
   if (x & STRIP_ANSI)
      *p++ = 'a';
   *p = 0;
   return s;
}

char *stripmaskname(int x)
{
   static char s[161];
   s[0] = 0;
   if (x & STRIP_BOLD)
      strcat(s, "bold, ");
   if (x & STRIP_COLOR)
      strcat(s, "color, ");
   if (x & STRIP_REV)
      strcat(s, "reverse, ");
   if (x & STRIP_UNDER)
      strcat(s, "underline, ");
   if (x & STRIP_ANSI)
      strcat(s, "ansi, ");
   if (!s[0])
      strcpy(s, "none, ");
   s[strlen(s) - 2] = 0;
   return s;
}

void cmd_strip PROTO2(int, idx, char *, par)
{
   char nick[512], s[2], s1[512];
   int dest = 0, i, pls, md, ok = 0, atr;
   if (!par[0]) {
     dprintf(idx,"[1;37mYour current strip settings are:[0m [1;33m%s (%s)[0m\n",
     	      stripmasktype(dcc[idx].u.chat->strip_flags),
	      stripmaskname(dcc[idx].u.chat->strip_flags));
      return;
   }
   strcpy(s1, par);
   split(nick, par);
   atr = get_attr_handle(dcc[idx].nick);
   if ((nick[0]) && (nick[0] != '+') && (nick[0] != '-') &&
       (atr & USER_MASTER)) {
      for (i = 0; i < dcc_total; i++)
	 if ((strcasecmp(nick, dcc[i].nick) == 0) && (dcc[i].type == DCC_CHAT) &&
	     (!ok)) {
	    ok = 1;
	    dest = i;
	 }
      if (!ok) {
	 dprintf(idx,"[1;31mNo such user on the party line![0m\n");
	 return;
      }
      nick[0] = 0;
   } else
      dest = idx;
   if (!nick[0])
      nsplit(nick, par);
   while (nick[0]) {
      pls = 1;
      if ((nick[0] != '+') && (nick[0] != '-'))
	 dcc[dest].u.chat->strip_flags = 0;
      for (i = 0; i < strlen(nick); i++) {
	 if (nick[i] == '+')
	    pls = 1;
	 else if (nick[i] == '-')
	    pls = (-1);
	 else {
	    s[0] = nick[i];
	    s[1] = 0;
	    md = stripmodes(s);
	    if (pls == 1)
	       dcc[dest].u.chat->strip_flags |= md;
	    else
	       dcc[dest].u.chat->strip_flags &= ~md;
	 }
      }
      nsplit(nick, par);
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m strip %s",dcc[idx].nick, s1);
   if (dest == idx) {
     dprintf(idx,"[1;33mYour strip settings are:[0m [1;37m%s (%s)[0m\n", 
     	      stripmasktype(dcc[idx].u.chat->strip_flags),
	      stripmaskname(dcc[idx].u.chat->strip_flags));
   } else {
     dprintf(idx,"[1;33mStrip settings for %s:[0m [1;32m%s (%s)[0m\n",dcc[dest].nick, 
           stripmasktype(dcc[dest].u.chat->strip_flags),
	      stripmaskname(dcc[dest].u.chat->strip_flags));
      dprintf(idx,"[1;31m%s set your strip settings to:[0m [1;32m%s (%s)[0m\n",dcc[idx].nick, 
      	      stripmasktype(dcc[dest].u.chat->strip_flags),
	      stripmaskname(dcc[dest].u.chat->strip_flags));
   }
}

void cmd_stick_yn(idx, par, yn)
int idx;
char *par;
int yn;
{
   int i, j;
   struct chanset_t *chan;
   char s[UHOSTLEN];
   if (!par[0]) {
      dprintf(idx,"[1;34mUsage: %sstick <ban>[0m\n",yn?"":"un");
      return;
   }
   i = setsticky_ban(par, yn);
   if (i > 0) {
       putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m %sstick %s",dcc[idx].nick,yn?"":"un",par);
       dprintf(idx,"[1;34m%stuck: %s[0m\n",yn?"S":"Uns",par);
      return;
   }
   /* channel-specific ban? */
   chan = findchan(dcc[idx].u.chat->con_chan);
   if (chan == NULL) {
      dprintf(idx,"[1;34mInvalid console channel.[0m\n");
      return;
   }
   if (atoi(par))
      sprintf(s, "%d", i + atoi(par));
   else
      strcpy(s, par);
   j = u_setsticky_ban(chan->bans, s, yn);
   if (j > 0) {
     putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m %sstick %s",dcc[idx].nick,yn?"":"un",s);
     dprintf(idx,"[1;34m%stuck: %s[0m\n",yn?"S":"Uns",s);
      return;
   }
   /* well fuxor then. */
    dprintf(idx,"[1;34mNo such ban.[0m\n");
}

void cmd_stick PROTO2(int, idx, char *, par)
{
   cmd_stick_yn(idx, par, 1);
}

void cmd_unstick PROTO2(int, idx, char *, par)
{
   cmd_stick_yn(idx, par, 0);
}

void cmd_pls_chrec PROTO2(int, idx, char *, par)
{
   char nick[512], chn[512];
   struct chanset_t *chan;
   struct userrec *u;
   struct chanuserrec *chanrec;
   context;
   if (!par[0]) {
      dprintf(idx, "[1;34mUsage:[0m [1;50m+chrec <User> [channel][0m\n");
      return;
   }
   nsplit(nick, par);
   u = get_user_by_handle(userlist, nick);
   if (u == NULL) {
      dprintf(idx, "[1;31mNo such user.[0m\n");
      return;
   }
   if (par[0] == 0)
      chan = findchan(dcc[idx].u.chat->con_chan);
   else {
      nsplit(chn, par);
      chan = findchan(chn);
   }
   if (chan == NULL) {
      dprintf(idx, "[1;31mNo such channel.[0m\n");
      return;
   }
   if (!(get_attr_handle(dcc[idx].nick) & USER_MASTER) &&
   !(get_chanattr_handle(dcc[idx].nick, chan->name) & CHANUSER_MASTER)) {
      dprintf(idx, "[1;33mYou have no permission to do that on %s.[0m\n",
	      chan->name);
      return;
   }
   chanrec = get_chanrec(u, chan->name);
   if (chanrec != NULL) {
      dprintf(idx, "[1;37mUser %s[0m [1;31malready has a channel record for %s.[0m\n", nick, chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m +chrec %s %s", dcc[idx].nick, nick, chan->name);
   add_chanrec(u, chan->name, 0, 0);
   dprintf(idx, "[1;33mAdded %s channel record for[0m [1;50m%s.[0m\n", chan->name, nick);
}

void cmd_mns_chrec PROTO2(int, idx, char *, par)
{
   char nick[512], chn[512];
   struct chanset_t *chan;
   struct userrec *u;
   struct chanuserrec *chanrec;
   context;
   if (!par[0]) {
      dprintf(idx, "[1;34mUsage:[0m [1;50m-chrec <User> [channel][0m\n");
      return;
   }
   nsplit(nick, par);
   u = get_user_by_handle(userlist, nick);
   if (u == NULL) {
      dprintf(idx, "[1;32mNo such user.[0m\n");
      return;
   }
   if (par[0] == 0)
      chan = findchan(dcc[idx].u.chat->con_chan);
   else {
      nsplit(chn, par);
      chan = findchan(chn);
   }
   if (chan == NULL) {
      dprintf(idx, "[1;31mNo such channel.[0m\n");
      return;
   }
   if (!(get_attr_handle(dcc[idx].nick) & USER_MASTER) &&
   !(get_chanattr_handle(dcc[idx].nick, chan->name) & CHANUSER_MASTER)) {
      dprintf(idx, "[1;37mYou have no permission to do that on %s.[0m\n",
	      chan->name);
      return;
   }
   chanrec = get_chanrec(u, chan->name);
   if (chanrec == NULL) {
      dprintf(idx, "[1;31mUser %s[0m [1;32mdoesn't have a channel record for %s.[0m\n", nick, chan->name);
      return;
   }
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m -chrec %s %s", dcc[idx].nick, nick, chan->name);
   del_chanrec(u, chan->name);
   dprintf(idx, "[1;37mRemoved %s channel record for[0m [1;50m%s.[0m\n", chan->name, nick);
}

void cmd_su PROTO2(int, idx, char *, par)
{
   int atr = get_attr_handle(dcc[idx].nick);
   if (strlen(par) == 0) {
      dprintf(idx, "[1;34mUsage:[0m [1;50msu <user>[0m\n");
      return;
   }
   if (!is_user(par)) {
      dprintf(idx, "[1;31mNo such user.[0m\n");
      return;
   }
   if (get_attr_handle(par) & USER_BOT) {
      dprintf(idx, "[1;32mWhy would you want to be a bawt... then again, why not?[0m\n");
      return;
   }
   if (pass_match_by_handle("-", par)) {
      dprintf(idx, "[1;37mNo password set for user. You may not .su to them[0m\n");
      return;
   }
   correct_handle(par);
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m su %s", dcc[idx].nick, par);
   if (!(atr & USER_OWNER) || (get_attr_handle(par) & USER_OWNER)) {
      tandout("part %s %s %d\n", botnetnick, dcc[idx].nick, dcc[idx].sock);
      chanout2(dcc[idx].u.chat->channel, "%s left this lifeless realm.\n",
	       dcc[idx].nick);
      context;
      /* store the old nick in the away section, for weenies who can't get
       * their password right ;) */
      if (dcc[idx].u.chat->away != NULL)
	 nfree(dcc[idx].u.chat->away);
      dcc[idx].u.chat->away = n_malloc(strlen(dcc[idx].nick) + 1, "dccutil.c", 9999);
      strcpy(dcc[idx].u.chat->away, dcc[idx].nick);
      strcpy(dcc[idx].nick, par);
      dprintf(idx, "[1;37mEnter password for %s[0m\n", par);
      dcc[idx].type = DCC_CHAT_PASS;
      return;
   }
   if (atr & USER_OWNER) {
      tandout("part %s %s %d\n", botnetnick, dcc[idx].nick, dcc[idx].sock);
      chanout2(dcc[idx].u.chat->channel, "%s left this lifeless realm.\n", dcc[idx].nick);
      context;
      dprintf(idx, "[1;33mSetting your username to %s.[0m\n", par);
      strcpy(dcc[idx].nick, par);
      tandout("join %s %s %d %c%d %s\n", botnetnick, dcc[idx].nick,
	      dcc[idx].u.chat->channel, geticon(idx), dcc[idx].sock,
	      dcc[idx].host);
      chanout2(dcc[idx].u.chat->channel, "%s has joined this lifeless realm.\n", dcc[idx].nick);
      context;
      return;
   }
}

#ifndef NO_IRC

void cmd_pls_chan PROTO2(int, idx, char *, par)
{
   char chname[512], null[] = "";
   if (!par[0]) {
      dprintf(idx, "[1;34mUsage:[0m [1;50m+chan <#channel>[0m\n");
      return;
   }
   nsplit(chname, par);
   if (findchan(chname) != NULL) {
      dprintf(idx, "[1;37mThat channel already exists![0m\n");
      return;
   }
   if (tcl_channel_add(0, chname, null) == TCL_ERROR)
      dprintf(idx, "[1;37mInvalid channel.[0m\n");
   else
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m +chan %s", dcc[idx].nick, chname);
}

void cmd_mns_chan PROTO2(int, idx, char *, par)
{
   char chname[512];
   struct chanset_t *chan;
   int i;
   if (!par[0]) {
      dprintf(idx, "[1;34mUsage:[0m [1;50m-chan <#channel>[0m\n");
      return;
   }
   nsplit(chname, par);
   chan = findchan(chname);
   if (chan == NULL) {
      dprintf(idx, "[1;31mThat channel doesnt exist![0m\n");
      return;
   }
   if (chan->stat & CHANSTATIC) {
      dprintf(idx, "[1;32mCannot remove %s, it is not a dynamic channel!.[0m\n",
	      chname);
      return;
   }
   clear_channel(chan, 0);
   freeuser(chan->bans);
   killchanset(chname);
   if (serv >= 0)
      mprintf(serv, "PART %s\n", chname);
   dprintf(idx, "[1;33mChannel %s[0m [1;32mremoved from the bot.[0m\n", chname);
   dprintf(idx, "[1;31mThis includes any channel specific bans you set.[0m\n");
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m -chan %s", dcc[idx].nick, chname);
   for (i = 0; i < dcc_total; i++)
      if ((dcc[i].type == DCC_CHAT)
	  && (strcasecmp(dcc[i].u.chat->con_chan, chname) == 0)) {
	 dprintf(i, "[1;33m%s[0m [1;32mis no longer a valid channel, changing your console to '*'[0m\n",
		 chname);
	 strcpy(dcc[i].u.chat->con_chan, "*");
      }
}

void cmd_chaninfo PROTO2(int, idx, char *, par)
{
   char chname[256], work[512];
   struct chanset_t *chan;
   if (!par[0]) {
      dprintf(idx, "[1;34mUsage:[0m [1;50mchaninfo <channel>[0m\n");
      return;
   }
   nsplit(chname, par);
   chan = findchan(chname);
   if (chan == NULL) {
      dprintf(idx, "[1;31mNo such channel defined.[0m\n");
      return;
   }
   if (chan->stat & CHANSTATIC)
      dprintf(idx, "[1;37mSettings for static channel[0m [1;32m%s[0m\n", chname);
   else
      dprintf(idx, "[1;37mSettings for dynamic channel[0m [1;32m%s[0m\n", chname);
   get_mode_protect(chan, work);
   dprintf(idx, "[1;31mProtect modes (chanmode):[0m [1;33m%s[0m\n", work[0] ? work : "[1;50mNone[0m");
   if (chan->idle_kick)
      dprintf(idx, "[1;32mIdle Kick after (idle-kick):[0m [1;31m%d[0m\n", chan->idle_kick);
   else
      dprintf(idx, "[1;32mIdle Kick after (idle-kick):[0m [1;50mDONT![0m\n");
   /* only bot owners can see/change these (they're TCL commands) */
   if (get_attr_handle(dcc[idx].nick) & USER_OWNER) {
      if (chan->need_op[0])
	 dprintf(idx, "[1;33mTo regain op's (need-op):[0m\n[1;32m%s[0m\n", chan->need_op);
      if (chan->need_invite[0])
	 dprintf(idx, "[1;33mTo get invite (need-invite):[0m\n[1;31m%s[0m\n", chan->need_invite);
      if (chan->need_key[0])
	 dprintf(idx, "[1;33mTo get key (need-key):[0m\n[1;37m%s[0m\n", chan->need_key);
      if (chan->need_unban[0])
	 dprintf(idx, "[1;33mIf Im banned (need-unban):[0m\n[1;50m%s[0m\n", chan->need_unban);
      if (chan->need_limit[0])
	 dprintf(idx, "[1;33mWhen channel full (need-limit):[0m\n[1;32m%s[0m\n", chan->need_limit);
   }
   dprintf(idx, "[1;31mOther modes:[0m\n");
   dprintf(idx, "    [1;32m%cclearbans[0m [1;37m %cenforcebans[0m [1;33m %cdynamicbans[0m [1;50m %cuserbans[0m\n",
	   (chan->stat & CHAN_CLEARBANS) ? '+' : '-',
	   (chan->stat & CHAN_ENFORCEBANS) ? '+' : '-',
	   (chan->stat & CHAN_DYNAMICBANS) ? '+' : '-',
	   (chan->stat & CHAN_NOUSERBANS) ? '-' : '+');
   dprintf(idx, "    [1;50m%cautoop[0m     [1;31m%cbitch[0m       [1;32m %cgreet[0m       [1;37m %cprotectops[0m\n",
	   (chan->stat & CHAN_OPONJOIN) ? '+' : '-',
	   (chan->stat & CHAN_BITCH) ? '+' : '-',
	   (chan->stat & CHAN_GREET) ? '+' : '-',
	   (chan->stat & CHAN_PROTECTOPS) ? '+' : '-');
   dprintf(idx, "   [1;31m %cstatuslog[0m [1;32m %cstopnethack[0m [1;37m %crevenge[0m    [1;33m %csecret[0m\n",
	   (chan->stat & CHAN_LOGSTATUS) ? '+' : '-',
	   (chan->stat & CHAN_STOPNETHACK) ? '+' : '-',
	   (chan->stat & CHAN_REVENGE) ? '+' : '-',
	   (chan->stat & CHAN_SECRET) ? '+' : '-');
   dprintf(idx, "    [1;32m%cshared[0m     [1;50m%cautovoice[0m\n", (chan->stat & CHAN_SHARED) ? '+' : '-',(chan->stat & CHAN_VOICEJOIN) ? '+' : '-');
   putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chaninfo %s", dcc[idx].nick, chname);
}

void cmd_chanset PROTO2(int, idx, char *, par)
{
   char chname[512], work[512], args[512], answer[512];
   char *list[2];
   struct chanset_t *chan;
   if (!par[0]) {
      dprintf(idx, "[1;34mUsage:[0m [1;50mchanset <#channel> <settings>[0m\n");
      return;
   }
   nsplit(chname, par);
   chan = findchan(chname);
   if (chan == NULL) {
      dprintf(idx, "[1;33mThat channel doesnt exist![0m\n");
      return;
   }
   nsplit(work, par);
   list[0] = work;
   list[1] = args;
   answer[0] = 0;
   while (work[0]) {
      if (work[0] == '+' || work[0] == '-' ||
	  (strcmp(work, "dont-idle-kick") == 0)) {
	 if (strcmp(&(work[1]), "shared") == 0) {
	    dprintf(idx, "[1;37mYou can't change shared settings on the fly[0m\n");
	 } else if (tcl_channel_modify(0, chan, 1, list) == TCL_OK) {
	    strcat(answer, work);
	    strcat(answer, " ");
	 } else
	    dprintf(idx, "[1;31mError trying to set %s for %s, invalid mode[0m\n",
		    work, chname);
	 nsplit(work, par);
	 continue;
      }
      /* the rest have an unknow amount of args, so assume the rest of the *
       * line is args. Woops nearly made a nasty little hole here :) we'll *
       * just ignore any non global +n's trying to set the need-commands   */
      if ((strncmp(work, "need-", 5) != 0) ||
	  (get_attr_handle(dcc[idx].nick) & USER_OWNER)) {
	 strcpy(args, par);
	 if (tcl_channel_modify(0, chan, 2, list) == TCL_OK) {
	    strcat(answer, work);
	    strcat(answer, " { ");
	    strcat(answer, par);
	    strcat(answer, " }");
	 } else
	    dprintf(idx, "[1;31mError trying to set %s for %s, invalid option[0m\n",
		    work, chname);
      }
      break;
   }
   if (answer[0]) {
      dprintf(idx, "[1;31mSuccessfully set modes[0m [1;50m{ %s }[0m [1;33mon %s.[0m\n",
	      answer, chname);
      if (chan->stat & CHANSTATIC)
	 dprintf(idx, "[1;37mChanges to %s are not permanent.[0m\n", chname);
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m chanset %s %s", dcc[idx].nick, chname, answer);
   }
}

void cmd_chansave PROTO2(int, idx, char *, par)
{
   if (!chanfile[0])
   dprintf(idx, "[1;37mNo channel saving file defined.[0m\n");   
   else
      dprintf(idx, "Saving all dynamic channel settings.\n");
   putlog(LOG_MISC, "*", "[1;31m#[0m[1;34m%s[0m[1;31m#[0m chansave", dcc[idx].nick);
   write_channels();
}

void cmd_chanload PROTO2(int, idx, char *, par)
{
   if (!chanfile[0])
      dprintf(idx, "[1;37mNo channel saving file defined.[0m\n");
   else
      dprintf(idx, "[1;37mReloading all dynamic channel settings.[0m\n");
   putlog(LOG_MISC, "*", "[1;31m#[0m[1;34m%s[0m[1;31m#[0m chanload", dcc[idx].nick);
   read_channels();
}

#endif				/* !NO_IRC */

void cmd_fixcodes PROTO2(int, idx, char *, par)
{
   if (dcc[idx].u.chat->status & STAT_ECHO) {
      dcc[idx].u.chat->status = STAT_TELNET;
      dprintf(idx, "Turned on telnet codes\n");
      putlog(LOG_CMDS, "*", "#%s# fixcodes (telnet on)", dcc[idx].nick);
      return;
   }
   if (dcc[idx].u.chat->status & STAT_TELNET) {
      dcc[idx].u.chat->status = STAT_ECHO;
      dprintf(idx, "Turned off telnet codes\n");
      putlog(LOG_CMDS, "*", "#%s# fixcodes (telnet off)", dcc[idx].nick);
      return;
   }
}

void cmd_page PROTO2(int, idx, char *, par)
{
   int a;
   if (!par[0]) {
      if (dcc[idx].u.chat->status & STAT_PAGE) {
	 dprintf(idx, "[1;31mCurrently paging outputs to[0m [1;50m%d lines.[0m\n",
		 dcc[idx].u.chat->max_line);
      } else
	dprintf(idx,"[1;34mYou dont have paging on.[0m\n"); 
      return;
   }
   a = atoi(par);
   if (strcasecmp(par, "off") == 0 || (a == 0 && par[0] == 0)) {
      dcc[idx].u.chat->status &= ~STAT_PAGE;
      dcc[idx].u.chat->max_line = 1;	/* flush_lines needs this */
      while (dcc[idx].u.chat->buffer != NULL)
	 flush_lines(idx);
      dprintf(idx,"[1;34mPaging turned off.[0m\n");
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m page off",dcc[idx].nick);
      return;
   }
   if (a > 0) {
      dprintf(idx,"[1;34mPaging turned on, stopping every %d lines.[0m\n", a);
      dcc[idx].u.chat->status |= STAT_PAGE;
      dcc[idx].u.chat->max_line = a;
      dcc[idx].u.chat->line_count = 0;
      dcc[idx].u.chat->current_lines = 0;
      putlog(LOG_CMDS,"*","[1;31m#[0m[1;34m%s[0m[1;31m#[0m page %d",dcc[idx].nick, a);
      return;
   }
   dprintf(idx,"[1;31mUsage: page <off or #>[0m\n");
   
}

#ifndef MODULES
void cmd_nomodules PROTO2(int,idx,char *,par) {
    dprintf(idx,"[1;33mModules not supported.[0m\n");
}
#endif
